import boto3
import time
import logging
import uuid

from figcli.commands.config_context import ConfigContext
from figcli.commands.config_factory import ConfigFactory
from figcli.commands.help_context import HelpContext
from figcli.commands.help_factory import HelpFactory
from figcli.commands.iam_context import IAMContext
from figcli.commands.iam_factory import IAMFactory
from figcli.commands.factory import Factory
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.run_env import RunEnv
from figcli.commands.figgy_context import FiggyContext
from figcli.svcs.kms import KmsSvc
from figcli.svcs.config import ConfigService
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.auth.provider.provider_factory import SessionProviderFactory
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.svcs.auth.session_manager import SessionManager
from figcli.data.dao.config import ConfigDao
from figcli.utils.utils import Utils
from figcli.config import *
from figcli.data.dao.ssm import SsmDao
from typing import Dict
from concurrent.futures import ThreadPoolExecutor, thread, as_completed

from figcli.views.rbac_limited_config import RBACLimitedConfigView
from threading import Lock
logger = logging.getLogger(__name__)


class CommandFactory(Factory):
    """
    Uses the provided FiggyContext (which contains details of args passed in, etc), and initializes a command
    factory of the appropriate type, and generates the appropriate command.
    """

    def __init__(self, context: FiggyContext, cli_defaults: CLIDefaults):
        self._id = uuid.uuid4()
        self._context = context
        self._utils = Utils(context.colors_enabled)
        self._cli_defaults = cli_defaults
        self._session_mgr = None
        self._session_provider = None
        self._env_session = None
        self._ssm = None
        self._config = None
        self._kms = None
        self._s3_rsc = None
        self._all_sessions = None
        self._config_svc = None
        self._cache_mgr = None
        self._rbac_config_view = None
        self.__env_lock = Lock()
        self.__mgr_lock = Lock()

    def __session_provider(self) -> SessionProvider:
        if not self._session_provider:
            self._session_provider = SessionProviderFactory(self._cli_defaults).instance()

        return self._session_provider

    def __session_manager(self):
        """
        Lazy load the session manager, only create a session if this command requires it.
        :return: 
        """
        with self.__mgr_lock:
            if not self._session_mgr:
                self._session_mgr = SessionManager(self._cli_defaults, self.__session_provider())

        return self._session_mgr

    def __env_session(self) -> boto3.session.Session:
        """
        Lazy load an ENV session object for the ENV selected in the FiggyContext
        :return: Hydrated session for the selected environment.
        """
        with self.__env_lock:
            if not self._env_session:
                self._env_session = self.__session_manager().get_session(
                    self._context.selected_role,
                    prompt=False)

        return self._env_session

    def __ssm(self) -> SsmDao:
        """
        Returns an SSMDao initialized with a session for the selected ENV based on FiggyContext
        """
        if not self._ssm:
            self._ssm = SsmDao(self.__env_session().client('ssm'))

        return self._ssm

    def __kms(self) -> KmsSvc:
        """
        Returns a hydrated KMS Service object based on these selected ENV
        """
        if not self._kms:
            self._kms: KmsSvc = KmsSvc(self.__env_session().client('kms'))

        return self._kms

    def __config(self) -> ConfigDao:
        """
        Returns a hydrated ConfigDao for the selected environment.
        """
        return ConfigDao(self.__env_session().resource('dynamodb'))

    def __s3_resource(self):
        """
        Returns a hydrated boto3 S3 Resource for the mgmt account.
        """
        if not self._s3_rsc:
            self._s3_rsc = self.__env_session().resource('s3')

        return self._s3_rsc

    def __all_sessions(self) -> Dict[str, boto3.session.Session]:
        """
        Populates a DICT containing boto sessions for all 4 environments (dev -> prod).
        """
        assumable_roles = self._cli_defaults.assumable_roles
        matching_roles = list(set([x for x in assumable_roles if x.role == self._context.role]))

        if not self._all_sessions and self._context.all_profiles:
            self._all_sessions: Dict[str, boto3.session.Session] = {}

            with ThreadPoolExecutor(max_workers=10) as pool:
                session_futures: Dict[str, thread] = {
                    role.role.full_name: pool.submit(self.__session_manager().get_session, role, prompt=False)
                    for role in matching_roles
                }

                for env, future in session_futures.items():
                    self._all_sessions[env] = future.result()
        else:
            self._all_sessions = None

        return self._all_sessions

    def __cache_mgr(self) -> CacheManager:
        """Builds a cache manager service for the specified resource."""
        if not self._cache_mgr:
            self._cache_mgr: CacheManager = CacheManager(Utils.get_first(self._context.resource))

        return self._cache_mgr

    def __config_service(self) -> ConfigService:
        """Returns a hydrated ConfigService."""
        if not self._config_svc:
            self._config_svc = ConfigService(self.__config(), self.__cache_mgr(), self._context.run_env)

        return self._config_svc

    def __rbac_config_view(self) -> RBACLimitedConfigView:
        if not self._rbac_config_view:
            self._rbac_config_view = RBACLimitedConfigView(self._context.role, self.__cache_mgr(),
                                                           self.__ssm(), self.__config_service())
        return self._rbac_config_view

    def __init_sessions(self):
        """
        Bootstraps sessions (blocking) before we do threaded lookups that require these sessions.
        """
        self.__session_manager().get_session(
                self._context.selected_role,
                prompt=False)

    def instance(self):
        """
        Get an instance of a particular command based on the FiggyContext provided into this factory.
        """
        factory: Factory = None
        start = time.time()

        if self._context.command in config_commands and self._context.resource == config:
            self.__init_sessions()
            context = ConfigContext(self._context.run_env, self._context.role, self._context.args, config,
                                    defaults=self._cli_defaults)


            futures = set()
            # Multiple threads to init resources saves 500 - 1000 MS
            with ThreadPoolExecutor(max_workers=5) as pool:
                futures.add(pool.submit(self._ssm))
                futures.add(pool.submit(self.__kms))
                futures.add(pool.submit(self.__s3_resource))

            for future in as_completed(futures):
                pass  # Force lazy init for all futures.

            factory = ConfigFactory(self._context.command, context, self.__ssm(), self.__config(), self.__kms(),
                                    self.__s3_resource(), self._context.colors_enabled, self.__rbac_config_view(),
                                    self.__session_manager())

        elif self._context.command in iam_commands and self._context.resource == iam:
            self.__init_sessions()
            context = IAMContext(self._context.run_env, self._context.role, self._context.colors_enabled, iam,
                                 defaults=self._cli_defaults)
            factory = IAMFactory(self._context.command, context, self.__env_session(),
                                 all_sessions=self.__all_sessions())
        elif self._context.find_matching_optional_arguments(help_commands) or self._context.resource in help_commands:
            optional_args = self._context.find_matching_optional_arguments(help_commands)
            context = HelpContext(self._context.resource, self._context.command, optional_args, self._context.run_env,
                                  defaults=self._cli_defaults)
            factory = HelpFactory(self._context.command, context)
        else:
            if self._context.command is None or self._context.resource:
                self._utils.error_exit(f"Proper {CLI_NAME} syntax is `{CLI_NAME} <ressource> <command>` "
                                       f"For example `{CLI_NAME} config get` Either resource or command were "
                                       f"not supplied.")
            else:
                self._utils.error_exit(
                    f"Command: {self._utils.get_first(self._context.command)} was not found in this version of figgy.")

        logger.info(f"Init completed in {time.time() - start} seconds.")
        return factory.instance()



from typing import Tuple, Optional

from botocore.exceptions import ClientError
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *


class Get(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, config_completer_init: WordCompleter,
                 colors_enabled: bool, config_context: ConfigContext):
        super().__init__(get, colors_enabled, config_context)
        self._ssm = ssm_init
        self._config_completer = config_completer_init
        self._utils = Utils(colors_enabled)

    def get_val_and_desc(self, key: str) -> Tuple[Optional[str], Optional[str]]:
        try:
            return self.get_param_and_desc(key)
        except ClientError as e:
            denied = "AccessDeniedException" == e.response['Error']['Code']
            if denied and "AWSKMS; Status Code: 400;" in e.response['Error']['Message']:
                print(f"\n{self.c.fg_rd}You do not have access to decrypt the value of: {key}{self.c.rs}")
            elif denied:
                print(f"\n{self.c.fg_rd}You do not have access to Parameter: {key}{self.c.rs}")
            else:
                raise
        return None, None

    def get_param_and_desc(self, key: str) -> Tuple[Optional[str], Optional[str]]:
        return self._ssm.get_parameter_with_description(key)

    def get(self, key: str) -> Optional[str]:
        return self._ssm.get_parameter(key)

    def _get_param(self):
        get_another = True

        while get_another:
            key = Input.input(f"PS Name: ",
                              completer=self._config_completer)
            if key:
                value, desc = self.get_val_and_desc(key)
                if value:
                    print(f"{self.c.fg_gr}Value: {self.c.rs}{value}")
                    desc = desc if desc else DESC_MISSING_TEXT
                    print(f"{self.c.fg_gr}Description: {self.c.rs}{desc}")
                else:
                    print(f"{self.c.fg_rd}Invalid PS Name specified.{self.c.rs}")
                get_another = Input.y_n_input(f"\nGet another?", default_yes=False, invalid_no=True)
                print()

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._get_param()



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.data.put import *
from figcli.test.cli.data.delete import DataDelete
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DataShare(FiggyTest):
    def __init__(self, extra_args=""):
        print(f"Testing `{CLI_NAME} config {Utils.get_first(share)} --env {DEFAULT_ENV}`")
        super().__init__(None, extra_args=extra_args)
        self._child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(share)} --env '
                                    f'{DEFAULT_ENV} --skip-upgrade {extra_args}',
                                    timeout=10, encoding='utf-8')

    def run(self):
        self.step("Adding parameters to share...")
        put = DataPut()
        put.add(data_param_1, data_param_1_val, data_param_1_desc, add_more=False)

        self.step(f"Testing share success from {figgy} to {automated_test_dest_1}")
        self.share(data_param_1, automated_test_dest_1, expect_failure=False, share_another=True)

        self.step("Testing share failure due to inaccessible namespace.")
        self.share(devops_param_1, automated_test_dest_1, expect_failure=True)

        delete = DataDelete()
        self.step("Testing failed deletion b/c source is a repl source.")
        delete.delete(data_param_1, delete_another=False, repl_source_delete=True)

        time.sleep(30)  # Wait for replication
        # print("Faking successful replication")
        # put.add(automated_test_dest_1, data_param_1_val, data_param_1_desc)

        delete = DataDelete(extra_args=self.extra_args)
        self.step("Testing successful deletion of a repl destination.")
        delete.delete(automated_test_dest_1, delete_another=False, repl_dest_delete=True)

    def share(self, source, destination, expect_failure=False, share_another=False):
        self.expect(".*Name you wish to share:.*")
        self.sendline(source)
        self.expect(".*destination of the shared value:.*")
        self.sendline(destination)

        if expect_failure:
            self.expect(".*ERROR:.*")
        else:
            self.expect(f".*{source}.*successfully shared.*Share another.*")
            if share_another:
                self.sendline('y')
            else:
                self.sendline('n')




from abc import ABC
import sys
from typing import List, Dict

from tabulate import tabulate

from figcli.commands.help_context import HelpContext
from figcli.commands.types.help import HelpCommand
from figcli.config import *
from figcli.config.style.terminal_factory import TerminalFactory
from figcli.utils.utils import Utils
from figcli.input.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.setup import FiggySetup
from figcli.svcs.auth.provider.provider_factory import SessionProviderFactory
from figcli.svcs.auth.provider.session_provider import SessionProvider


class Configure(HelpCommand, ABC):
    """
    Drives the --configure command
    """

    def __init__(self, help_context: HelpContext, figgy_setup: FiggySetup):
        super().__init__(configure, False, help_context)
        self._setup = figgy_setup

    def configure(self) -> CLIDefaults:
        """
        Orchestrates the --configure option. Writes selections to a defaults file in user's home dir.

        This default files stores the following information:

        user: User Name for SSO integrations
        role: The user's preferred default role if --role is not specified
        env: The user's preferred eefault environment if --env is not specified
        valid_roles: A list of roles the user has access to based on the returned SAML assertion from the SSO provider
        valid_envs: A list of environments the user has access to based on the returned SAML assertion
        assumable_roles: Maintains a mapping of accountId -> environment name -> role name so the we can authenticate
                         the user with the appropriate AWS accounts based on their returned SAML assertion.
        """
        Utils.wipe_vaults() or Utils.wipe_defaults() or Utils.wipe_config_cache()
        defaults: CLIDefaults = self._setup.get_defaults()
        defaults = self._setup.configure_auth(defaults)
        defaults = self._setup.configure_extras(defaults)
        self._setup.save_defaults(defaults)
        self.c = TerminalFactory(Utils.is_mac()).instance().get_colors()

        defaults = self._setup.configure_roles(current_defaults=defaults)
        defaults = self._setup.configure_preferences(defaults)
        defaults = self._setup.configure_figgy_defaults(defaults)
        self._setup.save_defaults(defaults)
        print(f"\n{self.c.fg_gr}Setup successful! Enjoy figgy!{self.c.rs}")
        return defaults

    def execute(self):
        self.configure()



#!/usr/bin/env python

import json
import time

import requests
from u2flib_host import u2f, exc, appid
from u2flib_host.constants import APDU_USE_NOT_SATISFIED

"""
Taken from: https://github.com/cevoaustralia/aws-google-auth/blob/master/aws_google_auth/u2f.py
We can't use aws_google_auth b/c of issues with homberw installations 
"""

def __appid_verifier__fetch_json(app_id):
    target = app_id
    while True:
        resp = requests.get(target, allow_redirects=False, verify=True)

        # If the server returns an HTTP redirect (status code 3xx) the
        # server must also send the header "FIDO-AppID-Redirect-Authorized:
        # true" and the client must verify the presence of such a header
        # before following the redirect. This protects against abuse of
        # open redirectors within the target domain by unauthorized
        # parties.
        if 300 <= resp.status_code < 400:
            if resp.headers.get('FIDO-AppID-Redirect-Authorized') != \
                    'true':
                raise ValueError('Redirect must set '
                                 'FIDO-AppID-Redirect-Authorized: true')
            target = resp.headers['location']
        else:
            return resp.json()


def __appid_verifier__valid_facets(app_id, facets):
    return facets


def u2f_auth(challenges, facet):
    devices = u2f.list_devices()
    for device in devices[:]:
        try:
            device.open()
        except:
            # Some U2F devices fail on the first attempt to open but
            # succeed on subsequent attempts. So retry once.
            try:
                device.open()
            except:
                devices.remove(device)

    try:
        prompted = False
        while devices:
            removed = []
            for device in devices:
                remove = True
                for challenge in challenges:
                    try:
                        return u2f.authenticate(device, json.dumps(challenge),
                                                facet)
                    except exc.APDUError as e:
                        if e.code == APDU_USE_NOT_SATISFIED:
                            remove = False
                            if not prompted:
                                print('Touch the flashing U2F device to '
                                      'authenticate...')
                                prompted = True
                        else:
                            pass
                    except exc.DeviceError:
                        pass
                if remove:
                    removed.append(device)
            devices = [d for d in devices if d not in removed]
            for d in removed:
                d.close()
            time.sleep(0.25)
    finally:
        for device in devices:
            device.close()
    raise RuntimeWarning("U2F Device Not Found")


appid.verifier.fetch_json = __appid_verifier__fetch_json
appid.verifier.valid_facets = __appid_verifier__valid_facets


from botocore.exceptions import ClientError
from prompt_toolkit import prompt

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.models.replication_config import ReplicationType, ReplicationConfig
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *
from figcli.config.style.style import FIGGY_STYLE


class Share(ConfigCommand):

    def __init__(self, ssm_init, config_init, config_completer_init, colors_enabled: bool,
                 config_context: ConfigContext):
        super().__init__(share, colors_enabled, config_context)

        self._ssm = ssm_init
        self._config = config_init
        self._config_completer = config_completer_init
        self._utils = Utils(colors_enabled)

    def _share_param(self):
        """
        Enables sharing of parameters from one namespace to the /app/service-name/replicated namespace.
        Args:
            run_env: Run Environment
        """

        source_name_msg = [
            (f'class:{self.c.bl}', 'Input the PS Name you wish to share: ')
        ]

        dest_name_msg = [
            (f'class:{self.c.bl}', 'Input the destination of the shared value: ')
        ]

        share_another = True
        while share_another:
            print()
            key = prompt(source_name_msg, completer=self._config_completer, style=FIGGY_STYLE)
            if re.match(f"{self.context.defaults.service_ns}/.*", key):
                print(
                    f"{self.c.fg_rd}The SOURCE of replication may not be from within the "
                    f"{self.context.defaults.service_ns}/ namespace.{self.c.rs}\n")
                continue

            dest = prompt(dest_name_msg, completer=self._config_completer, style=FIGGY_STYLE)
            key_value = None
            try:
                key_value = self._ssm.get_parameter(key)
            except ClientError as e:
                denied = "AccessDeniedException" == e.response['Error']['Code']
                if denied and "AWSKMS; Status Code: 400;" in e.response['Error']['Message']:
                    print(f"{self.c.fg_rd}You do not have access to decrypt the value of Name: {key}{self.c.rs}")
                elif denied:
                    print(f"{self.c.fg_rd}You do not have access to Name: {key}{self.c.rs}")
                else:
                    raise

                self._utils.validate(key_value is not None,
                                     "Either the Name you provided to share does not exist or you do not have the "
                                     "proper permissions to share the provided Name.")

            namespace = self._utils.parse_namespace(dest)

            repl_config = ReplicationConfig(dest, self.run_env,
                                            namespace, key, ReplicationType(REPL_TYPE_APP))
            self._config.put_config_repl(repl_config)
            print(f"{self.c.fg_gr}{key} successfully shared{self.c.rs}")
            to_continue = input(f"Share another? (Y/n): ")
            to_continue = to_continue if to_continue != '' else 'y'
            share_another = to_continue.lower() == "y"

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._share_param()



from abc import ABC, abstractmethod
from dataclasses import dataclass

from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.config import Config, CONFIG_OVERRIDE_FILE_PATH, SUPPORTED_OKTA_FACTOR_TYPES, FAKE_GOOGLE_IDP_ID, FAKE_GOOGLE_SP_ID, FAKE_OKTA_APP_LINK
from figcli.input.input import Input, List
from figcli.svcs.config_manager import ConfigManager
from figcli.utils.utils import Utils, Color

from figcli.models.defaults.provider import Provider
import re


class ProviderConfig(ABC):

    @staticmethod
    @abstractmethod
    def configure(mfa_enabled: bool = False) -> "ProviderConfig":
        pass


class ProviderConfigFactory:
    @staticmethod
    def instance(provider: Provider, mfa_enabled: bool = False) -> ProviderConfig:
        if provider is Provider.OKTA:
            return OktaProviderConfig.configure(mfa_enabled)
        elif provider is Provider.GOOGLE:
            return GoogleProviderConfig.configure(mfa_enabled)
        elif provider is Provider.AWS_BASTION:
            return BastionProviderConfig.configure(mfa_enabled)


@dataclass
class BastionProviderConfig(ProviderConfig):
    profile_name: str

    @staticmethod
    def get_profile():
        profile = input('Please input your aws profile linked to your credentials in your `bastion` account: ')
        Utils.stc_validate(profile != '', "You must input a valid profile name.")
        return profile

    @staticmethod
    def configure(mfa_enabled: bool = False) -> "BastionProviderConfig":
        config, c = ConfigManager(CONFIG_OVERRIDE_FILE_PATH), Utils.default_colors()
        profile = config.get_or_prompt(Config.Section.Bastion.PROFILE, BastionProviderConfig.get_profile)
        return BastionProviderConfig(profile_name=profile)

    @property
    def profile(self):
        return self.profile_name


@dataclass
class GoogleProviderConfig(ProviderConfig):
    """
        idp_id: str -> Google's assigned IDP Identifier for your google authentication.
                       This can be found in your IDP metadata.xml file (or elsewhere, probably)
        sp_id: str ->  Service Provider ID: from your configured SAML app.
    """
    idp_id: str
    sp_id: str

    @staticmethod
    def get_idp_id():
        idp_id = input('\nPlease input the Identity Provider ID associated with your Google Account: ')
        Utils.stc_is_valid_input(idp_id, "Identity Provider Id", True)
        return idp_id

    @staticmethod
    def get_sp_id():
        sp_id = input('\nPlease input the Service Provider ID associated with your Google SAML configuration (this is a '
            'number lke 123456789) : ')
        Utils.stc_is_valid_input(sp_id, "Service Provider Id", True)
        return sp_id

    @staticmethod
    def configure(mfa_enabled: bool = False) -> "GoogleProviderConfig":
        config, c = ConfigManager(CONFIG_OVERRIDE_FILE_PATH), Utils.default_colors()

        idp_id = config.get_property(Config.Section.Google.IDP_ID)
        sp_id = config.get_property(Config.Section.Google.SP_ID)

        if idp_id and idp_id != FAKE_GOOGLE_IDP_ID:
            idp_id = config.get_or_prompt(Config.Section.Google.IDP_ID, GoogleProviderConfig.get_idp_id)
        else:
            idp_id = config.get_or_prompt(Config.Section.Google.IDP_ID, GoogleProviderConfig.get_idp_id, force_prompt=True)

        if sp_id and sp_id != FAKE_GOOGLE_SP_ID:
            sp_id = config.get_or_prompt(Config.Section.Google.SP_ID, GoogleProviderConfig.get_sp_id)
        else:
            sp_id = config.get_or_prompt(Config.Section.Google.SP_ID, GoogleProviderConfig.get_sp_id, force_prompt=True)

        return GoogleProviderConfig(idp_id=idp_id, sp_id=sp_id)


@dataclass
class OktaProviderConfig(ProviderConfig):
    """
        app_link: str Application embed link for the company's AWS application in OKTA
                  Looks something like this: https://your-company.okta.com/home/amazon_aws/ASDF12351fg1/234'
    """
    app_link: str
    factor_type: str

    @property
    def base_url(self) -> str:
        result = re.match(r'^.*https://(.*\.com)/.*', self.app_link)
        if result:
            return result.group(1)
        else:
            Utils.stc_error_exit(f"Unable to parse base url from OKTA application link: {self.app_link}, are you "
                                 f"sure this link is valid?")

    @staticmethod
    def get_app_link() -> str:
        app_link = Input.input("Please input your OKTA AWS Application Embed Link. It's usually something like "
                         "'https://your-company.okta.com/home/amazon_aws/ASDF12351fg1/234': ")
        Utils.stc_is_valid_input(app_link, "OKTA AWS Application URL", True)
        return app_link

    @staticmethod
    def get_factor_type() -> str:
        factor_type = prompt(f"\nPlease select your OKTA MFA Factor type. Supported Types are "
                             f"{SUPPORTED_OKTA_FACTOR_TYPES}: ",
                             completer=WordCompleter(SUPPORTED_OKTA_FACTOR_TYPES))
        Utils.stc_validate(OktaProviderConfig.factor_type in SUPPORTED_OKTA_FACTOR_TYPES,
                           f"You must select a factor type from: {SUPPORTED_OKTA_FACTOR_TYPES}")
        return factor_type

    @staticmethod
    def configure(mfa_enabled: bool = False) -> "OktaProviderConfig":
        config, c = ConfigManager(CONFIG_OVERRIDE_FILE_PATH), Utils.default_colors()
        app_link = config.get_property(Config.Section.Google.IDP_ID)
        if app_link and app_link != FAKE_OKTA_APP_LINK:
            app_link = config.get_or_prompt(Config.Section.Okta.APP_LINK, OktaProviderConfig.get_app_link)
        else:
            app_link = config.get_or_prompt(Config.Section.Okta.APP_LINK, OktaProviderConfig.get_app_link, force_prompt=True)

        if mfa_enabled:
            factor_type = config.get_or_prompt(Config.Section.Okta.FACTOR_TYPE, OktaProviderConfig.get_factor_type)
        else:
            factor_type = None

        return OktaProviderConfig(app_link=app_link, factor_type=factor_type)



from abc import ABC, abstractmethod
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.sso.okta.okta_session import OktaSession


class OktaAuth(ABC):

    def __init__(self, defaults: CLIDefaults):
        self._defaults = defaults

    @property
    def app_link(self):
        return self._defaults.provider_config.app_link

    @property
    def base_url(self):
        return self._defaults.provider_config.base_url

    def factor_type(self):
        return self._defaults.provider_config.factor_type

    @abstractmethod
    def get_session(self) -> OktaSession:
        pass



import logging
from abc import ABC, abstractmethod
from typing import List, Tuple

import boto3
from botocore.exceptions import ClientError

from figcli.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils
from threading import Lock
log = logging.getLogger(__name__)


class SessionProvider(ABC):
    def __init__(self, defaults: CLIDefaults):
        self._defaults = defaults
        self._valid_tokens = set()
        self._lock = Lock()

    @Utils.retry
    @Utils.trace
    def _is_valid_session(self, session: boto3.Session):
        """Tests whether a cached session is valid or not."""
        with self._lock:
            token = session.get_credentials().get_frozen_credentials().token
            if token in self._valid_tokens:
                log.info(f"Session with token: {token} was already validated.")
                return True
            else:
                try:
                    log.info(f"Testing session with token: {token}")
                    sts = session.client('sts')
                    sts.get_caller_identity()
                    self._valid_tokens.add(token)
                    return True
                except ClientError:
                    return False

    @abstractmethod
    def get_session(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) -> boto3.Session:
        pass

    def get_session_and_role(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) \
            -> Tuple[boto3.Session, AssumableRole]:
        return self.get_session(assumable_role, prompt, exit_on_fail), assumable_role


    @abstractmethod
    def cleanup_session_cache(self):
        pass

    def _get_user(self, prompt: bool) -> str:
        """
        Get the user either from cache, or prompt the user.

        Returns: str -> username
        """

        defaults = self._defaults
        if defaults is not None and not prompt:
            return defaults.user
        else:
            return Input.get_user(provider=self._defaults.provider.name)

    def _get_password(self, user_name, prompt: bool, save: bool = False) -> str:
        """
        Get the password either from keyring, or prompt the user.

        Returns: str -> password
        """

        password = SecretsManager.get_password(user_name)
        reset_password = not password

        if reset_password or prompt:
            password = Input.get_password(provider=self._defaults.provider.name)
            if reset_password or save:
                SecretsManager.set_password(user_name, password)

        return password



from abc import ABC

from figcli.commands.command_context import CommandContext
from figcli.commands.iam_context import IAMContext
from figcli.commands.types.command import Command


class IAMCommand(Command, ABC):
    """
    Config command class from which all other config command classes inherit.
    """

    def __init__(self, command_type: frozenset, context: IAMContext):
        super().__init__(command_type, context.colors_enabled, CommandContext(context.run_env, context.resource,
                                                                              context.defaults))
        self.role = context.role



from boto3.dynamodb.conditions import Key, Attr
from decimal import *
from figcli.config import *
from figcli.data.models.config_item import ConfigItem
from figcli.models.replication_config import ReplicationConfig, ReplicationType
from figcli.models.restore_config import RestoreConfig
from figcli.models.run_env import RunEnv
from figcli.models.audit_log import AuditLog
from figcli.models.parameter_history import ParameterHistory
from figcli.models.parameter_store_history import PSHistory
from typing import Callable, Dict, List, Any, Set
import json
import logging
import datetime
import time

log = logging.getLogger(__name__)


class ConfigDao:
    """
    This DAO executes queries against various DynamoDB tables required for our config management system's operation.
    1) Config replication table
    2) Audit Table
    3) Cache table.
    """

    def __init__(self, dynamo_resource):
        self._dynamo_resource = dynamo_resource
        self._config_repl_table = self._dynamo_resource.Table(REPL_TABLE_NAME)
        self._audit_table = self._dynamo_resource.Table(AUDIT_TABLE_NAME)
        self._cache_table = self._dynamo_resource.Table(CACHE_TABLE_NAME)

    def delete_config(self, destination: str, run_env: RunEnv) -> None:
        """
        Deletes a Replication configuration from the DB
        Args:
            destination: str -> /path/to/configuration/destination
            run_env: RunEnv -> run_env
        """
        self._config_repl_table.delete_item(
            Key={REPL_DEST_KEY_NAME: destination, REPL_RUN_ENV_KEY_NAME: run_env.env}
        )

    def get_parameter_restore_details(self, ps_name: str, start_key: str = None) -> List[RestoreConfig]:
        """
        :param ps_name:  str -> parameter store key name
        :param start_key: str -> used for recursive lookups on table scans.
        :return:
            List of parameter name + value + description + type
        """
        results = []
        filter_exp = Key(AUDIT_PARAMETER_KEY_NAME).eq(ps_name)
        if start_key:
            result = self._audit_table.scan(FilterExpression=filter_exp, ExclusiveStartKey=start_key)
        else:
            result = self._audit_table.scan(FilterExpression=filter_exp, )

        items: List = result["Items"] if result["Items"] else []

        # Remove items from list where action != "PutParameter"
        items = list(filter(lambda x: x["action"] == "PutParameter", items))

        results = sorted(RestoreConfig.convert_to_model(items), key=lambda x: x.ps_time, reverse=True)

        if 'LastEvaluatedKey' in result:
            results = results + self.get_parameter_restore_details(ps_name, start_key=result['LastEvaluatedKey'])

        # Convert to RestoreConfig model then sort list chronologically by timestamp
        return results

    def get_all_parameter_history(self, ps_time: datetime.datetime, ps_prefix: str, start_key=None) -> List[Dict]:
        """
        Scans in DynamoDb only do 1MB at at time. For large tables, we need to tell dynamo to KEEP_SCANNING. After each
        result set is returned, we need to inform dynamo to keep scanning to find ALL results across the full table. This
        will continue going back to dynamo until we find the _full_ history of a parameter.
        Args:
            ps_time: Time up to which parameter history should be returned.
            ps_prefix: e.g. /shared/some/prefix - Prefix to query under
            start_key: LastEvaluatedKey returned in scan results. Lets you konw if there is more scanning that can be done.
        Returns: List[Dict] of items returned from AWS API
        """
        time_end = Decimal(ps_time.timestamp() * 1000)
        filter_exp = Key(AUDIT_TIME_KEY_NAME).lt(time_end) & Key(AUDIT_PARAMETER_KEY_NAME).begins_with(ps_prefix)
        results = []

        if start_key:
            result = self._audit_table.scan(FilterExpression=filter_exp, ExclusiveStartKey=start_key)
        else:
            result = self._audit_table.scan(FilterExpression=filter_exp)

        if 'LastEvaluatedKey' in result:
            results.append(self.get_all_parameter_history(ps_time, ps_prefix, start_key=result['LastEvaluatedKey']))

        results.append(result)  # Add results from this query to list of lists

        results_list = list(map(lambda x: x['Items'], results))  # extract all the items
        result['Items'] = [item for sublist in results_list for item in sublist]  # Merge sublists to one list

        return result

    def get_parameter_history_before_time(self, ps_time: datetime.datetime, ps_prefix: str) -> PSHistory:
        """
        Retrieves total parameter history for all parameters up until the datetime passed in under the provided prefix.
        Args:
            ps_time: Time up to which parameter history should be returned.
            ps_prefix: e.g. /shared/some/prefix - Prefix to query under

        Returns:

        """

        result = self.get_all_parameter_history(ps_time, ps_prefix)
        items: List = result.get('Items', [])
        restore_cfgs: List[RestoreConfig] = RestoreConfig.convert_to_model(items)

        ps_histories: Dict[str, ParameterHistory] = {}
        for cfg in restore_cfgs:
            if ps_histories.get(cfg.ps_name, None):
                ps_histories.get(cfg.ps_name).add(cfg)
            else:
                ps_histories[cfg.ps_name] = ParameterHistory.instance(cfg)

        return PSHistory(list(ps_histories.values()))

    def get_parameter_restore_range(self, ps_time: datetime.datetime, ps_prefix: str, start_key: str = None) \
            -> List[RestoreConfig]:
        """
        :param ps_time:  int -> datetime to query dynamo timestamp from
        :param ps_prefix: str -> parameter store prefix we will recursively restore from/to (e.g., /app/demo-time)
        :return:
            List of parameter name + value + description + type for given time range
        """
        time_end = Decimal(ps_time.timestamp() * 1000)

        filter_exp = Key(AUDIT_TIME_KEY_NAME).lt(time_end) & Key(AUDIT_PARAMETER_KEY_NAME).begins_with(ps_prefix) \
                     & Attr(AUDIT_ACTION_ATTR_NAME).eq(SSM_PUT)

        if start_key:
            result = self._audit_table.scan(FilterExpression=filter_exp, ExclusiveStartKey=start_key)
        else:
            result = self._audit_table.scan(FilterExpression=filter_exp)

        items: List = result.get('Items', [])
        items = list(filter(lambda x: x["action"] == "PutParameter", items))

        return sorted(RestoreConfig.convert_to_model(items), key=lambda x: x.ps_time, reverse=True)

    def get_all_configs(self, run_env: RunEnv, namespace: str, start_key: str = None) -> List[ReplicationConfig]:
        """
        Retrieves all replication configs from the database for a particular namespace
        Args:
            start_key: LastEvaluatedKey returned in scan results. Lets you konw if there is more scanning that can be done.
            run_env: RunEnv -> run_env
            namespace: namespace  - e.g. /app/demo-time/

        Returns:
            List of ReplicationConfigs that match the namespace.

        """
        filter_exp = Attr(REPL_NAMESPACE_ATTR_NAME).eq(namespace) & Key(
            REPL_RUN_ENV_KEY_NAME
        ).eq(run_env.env)

        if start_key:
            result = self._config_repl_table.scan(FilterExpression=filter_exp, ExclusiveStartKey=start_key)
        else:
            result = self._config_repl_table.scan(FilterExpression=filter_exp)

        items = result["Items"] if result["Items"] else None

        configs = []
        if items is not None:
            for item in items:
                repl_config = ReplicationConfig(
                    item[REPL_DEST_KEY_NAME],
                    RunEnv(item[REPL_RUN_ENV_KEY_NAME]),
                    item[REPL_NAMESPACE_ATTR_NAME],
                    item[REPL_SOURCE_ATTR_NAME],
                    ReplicationType(item[REPL_TYPE_ATTR_NAME]),
                )
                configs.append(repl_config)

        if 'LastEvaluatedKey' in result:
            configs = configs + self.get_all_configs(run_env, namespace, start_key=result['LastEvaluatedKey'])

        return configs

    def get_cfgs_by_src(self, source: str, run_env: RunEnv) -> List[ReplicationConfig]:
        """
        Args:
            source: Source to perform table scan by
            run_env: RunEnv to query for.

        Returns: A list of matching replication confgs.
        """

        filter_exp = Attr(REPL_SOURCE_ATTR_NAME).eq(source) & Key(
            REPL_RUN_ENV_KEY_NAME
        ).eq(run_env.env)
        result = self._config_repl_table.scan(FilterExpression=filter_exp)
        return self._map_results(result)

    def get_config_repl(self, destination: str, run_env: RunEnv) -> ReplicationConfig:
        """
        Lookup a replication config by destination / run_env
        Args:
            destination: str: /app/demo-time/replicated/destination/path
            run_env: RunEnvironment -> run_env

        Returns: Matching replication config, or None if none match.
        """

        filter_exp = Key(REPL_DEST_KEY_NAME).eq(destination) & Key(
            REPL_RUN_ENV_KEY_NAME
        ).eq(run_env.env)
        result = self._config_repl_table.query(KeyConditionExpression=filter_exp)

        if "Items" in result and len(result["Items"]) > 0:
            items = result["Items"][0]

            config = ReplicationConfig(
                items[REPL_DEST_KEY_NAME],
                RunEnv(items[REPL_RUN_ENV_KEY_NAME]),
                items[REPL_NAMESPACE_ATTR_NAME],
                items[REPL_SOURCE_ATTR_NAME],
                ReplicationType(items[REPL_TYPE_ATTR_NAME]),
                user=items[REPL_USER_ATTR_NAME],
            )
            return config
        else:
            return None

    def put_config_repl(self, config: ReplicationConfig) -> None:
        """
        Stores a replication configuration
        Args:
            config: ReplicationConfig -> a hydrated replication config object.
        """
        item = {
            REPL_DEST_KEY_NAME: config.destination,
            REPL_RUN_ENV_KEY_NAME: config.run_env,
        }

        for key in config.props:
            if (
                    key != REPL_DEST_KEY_NAME
                    and key != REPL_RUN_ENV_KEY_NAME
                    and not isinstance(config.props[key], float)
            ):
                item[key] = config.props[key]
            elif isinstance(config.props[key], float):
                item[key] = Decimal(f"{config.props[key]}")

        self._config_repl_table.put_item(Item=item)

    def _map_results(self, result: dict) -> List[ReplicationConfig]:
        """
        Takes a DDB Result object with a single result and maps it into a replication config
        Args:
            result: DDB boto3 result obj

        Returns: ReplicationConfig if result is found, else and empty list

        """
        repl_cfgs = []
        if "Items" in result and len(result["Items"]) > 0:
            for item in result["Items"]:
                config = ReplicationConfig(
                    item[REPL_DEST_KEY_NAME],
                    RunEnv(item[REPL_RUN_ENV_KEY_NAME]),
                    item[REPL_NAMESPACE_ATTR_NAME],
                    item[REPL_SOURCE_ATTR_NAME],
                    ReplicationType(item[REPL_TYPE_ATTR_NAME]),
                    user=item[REPL_USER_ATTR_NAME],
                )
                repl_cfgs.append(config)

        return repl_cfgs

    def get_audit_logs(self, ps_name: str) -> List[AuditLog]:
        """
        Args:
            ps_name: /path/to/parameter to query audit logs for.

        Returns: List[AuditLog]. Logs that match for the /ps/name in ParameterStore.
        """
        filter_exp = Key(AUDIT_PARAMETER_KEY_NAME).eq(ps_name)
        result = self._audit_table.query(KeyConditionExpression=filter_exp)
        items = result.get('Items', None)

        audit_logs: List[AuditLog] = []
        if items is not None:
            for item in items:
                log = AuditLog(item[AUDIT_PARAMETER_KEY_NAME], item[AUDIT_TIME_KEY_NAME],
                               item[AUDIT_ACTION_ATTR_NAME], item[AUDIT_USER_ATTR_NAME])
                audit_logs.append(log)

        return audit_logs

    def get_all_config_names(self, prefix: str = None, one_level: bool = False, start_key: str = None) -> Set[str]:
        """
        Retrieve all key names from the Dynamo DB config-cache table in each account. Much more efficient than
        querying SSM directly.
        Args:
            start_key: Optional: Is used for recursive paginated lookups to get the full data set. This should not
            be passed in by the user.
        Returns:

        """
        start_time = time.time()
        if start_key:
            log.info(f"Recursively scanning with start key: {start_key}")
            result = self._cache_table.scan(ExclusiveStartKey=start_key)
        else:
            result = self._cache_table.scan()

        configs: Set[str] = set()
        if "Items" in result and len(result['Items']) > 0:
            for item in result['Items']:
                name = item[CACHE_PARAMETER_KEY_NAME]
                configs.add(name)

        if prefix:
            configs = set(filter(lambda x: x.startswith(prefix), configs))

        if one_level:
            configs = set(filter(lambda x: len(x.split('/')) == len(prefix.split('/')) + 1, configs))

        if 'LastEvaluatedKey' in result:
            configs = configs | self.get_all_config_names(prefix, one_level, start_key=result['LastEvaluatedKey'])

        log.info(
            f"Returning config names from dynamo cache after: {time.time() - start_time} "
            f"seconds with {len(configs)} configs.")
        return configs

    def get_config_names_after(self, millis_since_epoch: int) -> Set[ConfigItem]:
        """
        Retrieve all key names from the Dynamo DB config-cache table in each account. Much more efficient than
        querying SSM directly.
        Args:
            millis_since_epoch: milliseconds in epoch to lookup config names from cache after
        Returns: Set[str] -> configs that have been added to cache table after millis_since_epoch
        """

        request = {
            'attribute_names': {
                '#n1': CACHE_LAST_UPDATED_KEY_NAME
            },
            'attribute_values': {
                ':v1': {'N': millis_since_epoch}
            },
            'expression': Attr(CACHE_LAST_UPDATED_KEY_NAME).gt(millis_since_epoch)
        }

        start_time = time.time()
        response = self._cache_table.scan(
            FilterExpression=request['expression']
        )

        items = response.get('Items', [])

        while 'LastEvaluatedKey' in response:
            response = self._cache_table.scan(ExclusiveStartKey=response['LastEvaluatedKey'])
            items = items + response.get('Items', [])

        configs: Set[ConfigItem] = set()
        for item in items:
            configs.add(ConfigItem.from_dict(item))

        log.info(f"Returning {len(configs)} parameter names from dynamo cache after time: [{millis_since_epoch}] in "
                 f"{time.time() - start_time} seconds.")

        return configs

    def put_in_config_cache(self, name):
        item = {
            CACHE_PARAMETER_KEY_NAME: name
        }

        self._cache_table.put_item(Item=item)



from pathlib import Path

VERSION = '0.0.33'
CLI_NAME = 'figgy'
PROJECT_NAME = 'figgy'

# Figgy URLS
FIGGY_GITHUB = "https://github.com/mancej/figgy"
FIGGY_OBS_API_BASE_URL = "https://api.figgy.dev"

# Paths
FIGGY_ERROR_REPORTING_URL = f"{FIGGY_OBS_API_BASE_URL}/v1/log-error"
FIGGY_GET_VERSION_URL = f"{FIGGY_OBS_API_BASE_URL}/v1/version"
FIGGY_LOG_METRICS_URL = f"{FIGGY_OBS_API_BASE_URL}/v1/log-metrics"

# Role names are assumed to be prefixed with `figgy-` - Users may override by setting the below ENV variable.
FIGGY_ROLE_NAME_PREFIX = 'figgy-'
FIGGY_ROLE_PREFIX_OVERRIDE_ENV = 'FIGGY_ROLE_PREFIX_OVERRIDE'

# TODO: UPDATE THIS

# Table-specific Constants
REPL_TABLE_NAME = "figgy-config-replication"
REPL_DEST_KEY_NAME = "destination"
REPL_RUN_ENV_KEY_NAME = "run_env"
REPL_SOURCE_ATTR_NAME = "source"
REPL_NAMESPACE_ATTR_NAME = "namespace"
REPL_TYPE_ATTR_NAME = "type"
REPL_TYPE_APP = "app"
REPL_TYPE_MERGE = "merge"
REPL_USER_ATTR_NAME = "user"

AUDIT_TABLE_NAME = "figgy-config-auditor"
AUDIT_PARAMETER_KEY_NAME = "parameter_name"
AUDIT_TIME_KEY_NAME = "time"
AUDIT_ACTION_ATTR_NAME = "action"
AUDIT_USER_ATTR_NAME = "user"

CACHE_TABLE_NAME = 'figgy-config-cache'
CACHE_PARAMETER_KEY_NAME = "parameter_name"
CACHE_LAST_UPDATED_KEY_NAME = "last_updated"
CACHE_STATE_ATTR_NAME = 'state'
AUDIT_PARAMETER_ATTR_TYPE = "type"
AUDIT_PARAMETER_ATTR_VERSION = "version"
AUDIT_PARAMETER_ATTR_DESCRIPTION = "description"
AUDIT_PARAMETER_ATTR_VALUE = "value"
AUDIT_PARAMETER_ATTR_KEY_ID = "key_id"

# Merge Key constants
MERGE_KEY_PREFIX = "${"
MERGE_KEY_SUFFIX = "}"

# SSM Constants
SSM_SECURE_STRING = "SecureString"
SSM_STRING = "String"
SSM_PUT = 'PutParameter'
SSM_DELETE = 'DeleteParameter'

# Other PS Config constants
DEPLOY_GROUPS_PS_PREFIX = '/shared/deploy-groups/'

# figgy.json json keys
REPLICATION_KEY = 'replicate_figs'
MERGE_KEY = 'merged_figs'
CONFIG_KEY = 'app_figs'
SHARED_KEY = 'shared_figs'
REPOSITORY_KEY = "repositories"
IMPORTS_KEY = 'import'
OPTIONAL_NAMESPACE = 'twig'
REPL_FROM_KEY = 'replicate_from'
SOURCE_NS_KEY = 'source_twig'
PARAMETERS_KEY = 'parameters'
SERVICE_KEY = 'service'
PLUGIN_KEY = 'plugins'

# Config paths
PS_FIGGY_ACCOUNTS_PREFIX = '/figgy/accounts/'
PS_FIGGY_DEFAULT_SERVICE_NS_PATH = '/figgy/defaults/service-namespace'

# Replication Types:
repl_types = [REPL_TYPE_APP, REPL_TYPE_MERGE]

# File names / paths
HOME = str(Path.home())
DECRYPTER_S3_PATH_PREFIX = f'figgy/decrypt/'
AWS_CREDENTIALS_FILE_PATH = f'{HOME}/.aws/credentials'
AWS_CONFIG_FILE_PATH = f'{HOME}/.aws/config'
CACHE_OTHER_DIR = f'{HOME}/.figgy/cache/other'
DEFAULT_INSTALL_PATH = '/usr/local/bin/figgy'
ERROR_LOG_DIR = f'{HOME}/.figgy/errors'
CONFIG_OVERRIDE_FILE_PATH = f'{HOME}/.figgy/config'
DEFAULTS_FILE_CACHE_PATH = f'{CACHE_OTHER_DIR}/defaults.json'
CONFIG_CACHE_FILE_PATH = f'{CACHE_OTHER_DIR}/config-cache.json'
STS_SESSION_CACHE_PATH = f"{HOME}/.figgy/vault/sts/sessions"
SAML_SESSION_CACHE_PATH = f"{HOME}/.figgy/vault/sso/saml"
OKTA_SESSION_CACHE_PATH = f"{HOME}/.figgy/cache/okta/session"
GOOGLE_SESSION_CACHE_PATH = f"{HOME}/.figgy/cache/google/session"

# Defaults file keys
DEFAULTS_ROLE_KEY = 'role'
DEFAULTS_ENV_KEY = 'default_env'
DEFAULTS_COLORS_ENABLED_KEY = 'colors'
DEFAULTS_USER_KEY = 'user'
DEFAULTS_PROVIDER_KEY = 'provider'
DEFAULTS_PROFILE_KEY = 'profile'
DEFAULTS_REGION_KEY = 'region'
MFA_SERIAL_KEY = 'mfa_serial'
DEFAULTS_KEY = 'defaults'

# Cache File keys
OKTA_SESSION_TOKEN_CACHE_KEY = 'session_token'
OKTA_SESSION_ID_CACHE_KEY = 'session_id'

# Plaform Constants
LINUX = "Linux"
MAC = "Darwin"
WINDOWS = "Windows"
ROOT_USER = "root"

# Build configs
OVERRIDE_KEYRING_ENV_VAR = "OVERRIDE_KEYRING"
ONE_WEEK_SECONDS = 60 * 60 * 24 * 7

# Figgy Sandbox
SANDBOX_ROLES = ['dev', 'devops', 'sre', 'data', 'dba']
GET_SANDBOX_CREDS_URL = "https://q39v8f3u13.execute-api.us-east-1.amazonaws.com/sandbox-bastion/v1/get-credentials"
FIGGY_SANDBOX_REGION = 'us-east-1'
FIGGY_SANDBOX_PROFILE = 'figgy-sandbox'
DISABLE_KEYRING = 'disable-keyring'

# Guaranteed Namespaces
shared_ns = '/shared'
figgy_ns = '/figgy'

# PS PATHS:
ACCOUNT_ID_PATH = f'{figgy_ns}/account_id'

# Vault
FIGGY_VAULT_FILES = [OKTA_SESSION_CACHE_PATH, GOOGLE_SESSION_CACHE_PATH, STS_SESSION_CACHE_PATH, SAML_SESSION_CACHE_PATH]

# Environment Variables
APP_NS_OVERRIDE = 'FIGGY_APP_TREE_OVERRIDE'
FIGGY_DISABLE_KEYRING = 'FIGGY_DISABLE_KEYRING'  # used for automated tests.

# Keychain
KEYCHAIN_ENCRYPTION_KEY = 'figgy-encryption-key'

# This is only used for temporary sandbox sessions. This reduces user friction when experimenting by not having to interact
# with authenticating their OS Keychain.
DEFAULT_ENCRYPTION_KEY = 'wX1C0nK1glfzaWQU8SKukdS7XZgYlAMW5ueb_V3cfSE='

# Default paths to search for figgy.json in
DEFAULT_FIGGY_JSON_PATHS = ['figgy.json', 'figgy/figgy.json', 'config/figgy.json', '_figgy/figgy.json', '.figgy/figgy.json']


from typing import Optional

from figcli.config import *
from abc import ABC, abstractmethod
from figcli.commands.config_context import ConfigContext
from figcli.commands.command_context import CommandContext
from figcli.commands.types.command import Command
from figcli.models.defaults.defaults import CLIDefaults


class ConfigCommand(Command, ABC):
    """
    Config command class from which all other config command classes inherit.
    """

    def __init__(self, command_type: frozenset, colors_enabled: bool, context: ConfigContext):
        super().__init__(command_type, colors_enabled, context)
        self.role = context.role



from typing import Optional

from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.run_env import RunEnv


class CommandContext:
    """
    All commands, regardless of resource type, will need to know what RunEnvironment they are operating on. That is the
    purpose of this command context. Similar properties would be added here.
    """
    def __init__(self, run_env: RunEnv, resource: frozenset, defaults: Optional[CLIDefaults]):
        self.run_env = run_env  # type: RunEnv
        self.resource = resource
        self.defaults = defaults



import os

from figcli.config.style.terms.default import DefaultTerm
from figcli.config.style.terms.iterm import iTerm
from figcli.config.style.terms.apple import AppleTerminal


class TerminalFactory:
    TRUE_COLOR = 'truecolor'

    def __init__(self, colors_enabled: bool):
        self.term = os.environ.get('TERM')
        self.term_program = os.environ.get('TERM_PROGRAM')
        self.color_term = os.environ.get('COLORTERM')
        self.colors_enabled = colors_enabled

    def instance(self):
        if self.term_program == iTerm.TERM_PROGRAM or self.color_term == 'truecolor':
            return iTerm(self.colors_enabled)
        elif self.term_program == AppleTerminal.TERM_PROGRAM:
            return AppleTerminal(self.colors_enabled)
        else:
            return DefaultTerm(self.colors_enabled)



from abc import abstractmethod, ABC

from figcli.config.style.color import Color


class Term(ABC):
    TYPE = 'default'

    def __init__(self, colors_enabled: bool):
        self.colors_enabled = colors_enabled

    @abstractmethod
    def get_colors(self) -> Color:
        pass



import logging
import json
import os
import jsonpickle
from typing import Dict, Any, Union, Set, Tuple, List, Callable, FrozenSet
from figcli.config import *
from figcli.svcs.vault import FiggyVault
from figcli.utils.utils import Utils
from pathlib import Path

log = logging.getLogger(__name__)

EMPTY_CACHE_STR = 'CACHE_EMPTY'


def write_empty_cache_file(path: str):
    with open(path, "w") as cache:
        cache.write(EMPTY_CACHE_STR)
    log.info(f"Cache written at path: {path}")


def prime_cache(function):
    def inner(self, *args, **kwargs):
        if not os.path.exists(self._cache_file):
            log.info("Cache file doesn't exist. Priming cache file.")
            write_empty_cache_file(self._cache_file)
            log.info("Cache primed.")
        return function(self, *args, **kwargs)

    return inner


def wipe_bad_cache(function):
    def inner(self, *args, **kwargs):
        try:
            function(self, *args, **kwargs)
        except json.JSONDecodeError as e:
            log.info("Cache file has been corrupted, recreating.")
            os.remove(self._cache_file)
            write_empty_cache_file(self._cache_file)

        return function(self, *args, **kwargs)

    return inner


class CacheManager:
    """
    Manages a local cache file to enhance figgy performance
    """

    _STORE_KEY = 'cache'
    _LAST_WRITE_KEY = 'last_write'
    _LAST_REFRESH_KEY = 'last_refresh'
    DEFAULT_REFRESH_INTERVAL = 60 * 60 * 24 * 7 * 1000  # 1 week in MS

    def __init__(self, cache_name: Union[str, FrozenSet, None] = "default", file_override: Union[str, None] = None,
                 vault: FiggyVault = None):
        cache_name = Utils.get_first(cache_name) if isinstance(cache_name, frozenset) else cache_name
        self._cache_file: str = f'{CACHE_OTHER_DIR}/{cache_name}-cache.json' if not file_override else file_override
        os.makedirs(CACHE_OTHER_DIR, exist_ok=True)

        if file_override:
            os.makedirs("/".join(file_override.split('/')[:-1]), exist_ok=True)

        self.vault = vault

    def __decrypt(self, data: bytes) -> str:
        """
        Decrypt the contents of a file, except when the file has been primed and we don't know if this file is going
        to be encrypted or decrypted eventually. If the file's contents are the primed contents, return an empty
        pickled json object and avoid decryption.

        :param data: data read from the file to decrypt
        :return: decrypted data if it was decrypted, or just the data if there is no need to decrypt
        """

        if data == bytes(EMPTY_CACHE_STR, 'utf-8'):
            return jsonpickle.encode({})

        if self.vault:
            data = self.vault.decrypt(data).decode('utf-8')
            return data
        else:
            return data.decode('utf-8')

    def __encrypt(self, data: str) -> bytes:
        """
        Encrypt and store this data in the cache if this cache has an assigned cryptography vault.
        :param data: data to encrypt
        :return: encrypted bytes
        """

        if self.vault:
            return self.vault.encrypt(bytes(data, 'utf-8'))
        else:
            return bytes(data, 'utf-8')

    def __read(self) -> str:
        with open(self._cache_file, 'rb') as cache:
            return self.__decrypt(cache.read())

    def __write(self, data: str):
        with open(self._cache_file, 'wb') as cache:
            cache.write(self.__encrypt(data))

    def get_val_or_refresh(self, cache_key: str, refresher: Callable, *args, max_age: int = DEFAULT_REFRESH_INTERVAL) \
            -> Any:
        last_write, val = self.get_or_refresh(cache_key, refresher, *args, max_age=max_age)
        return val

    @prime_cache
    @wipe_bad_cache
    def get_or_refresh(self, cache_key: str, refresher: Callable, *args, max_age: int = DEFAULT_REFRESH_INTERVAL) \
            -> Tuple[int, Any]:

        if not isinstance(max_age, int):
            log.warning(f"Received an invalid max_age of : {max_age}. Defaulting to"
                        f" {CacheManager.DEFAULT_REFRESH_INTERVAL}")

            max_age = CacheManager.DEFAULT_REFRESH_INTERVAL

        last_write, val = self.get(cache_key)
        if Utils.millis_since_epoch() - last_write > max_age or not val:
            new_val = refresher(*args)
            self.write(cache_key, new_val)
            return Utils.millis_since_epoch(), new_val
        else:
            return last_write, val

    @prime_cache
    @wipe_bad_cache
    def last_refresh(self, cache_key: str) -> int:
        """
        Return the last time the cache at this key was totally overwritten (and not just appended to)
        :param cache_key: key
        :return: int - millis since epoch - the last time this cache key was totally overwritten by the write method.
        """

        cache_string = self.__read()
        contents: Dict = jsonpickle.decode(cache_string)

        return contents.get(cache_key, {}).get(self._LAST_REFRESH_KEY, 0)

    @prime_cache
    @wipe_bad_cache
    def write(self, cache_key: str, object: Any) -> None:
        """
        Write an object to a local cache file. This will overwrite any existing values.
        :param cache_key: str - The key to write your object to local cache under. You will use this to lookup from
                the cache later.
        :param object: Any - Object to write, must be serializable by `json` library
        """
        try:
            object = list(object) if isinstance(object, Set) else object  # set is not json serializable
            contents: Dict = jsonpickle.decode(self.__read())
            contents[cache_key] = {
                self._STORE_KEY: object,
                self._LAST_WRITE_KEY: Utils.millis_since_epoch(),
                self._LAST_REFRESH_KEY: Utils.millis_since_epoch()
            }

            self.__write(jsonpickle.dumps(contents))
        except Exception as e:
            print(f"Error writing to cache key: {cache_key}: {e}")

    @prime_cache
    @wipe_bad_cache
    def get(self, cache_key: str, default=None) -> Tuple[int, Any]:
        """
        Retrieve an item from the cache.
        :param cache_key: Key to return from the cache.
        :param default: Default value to return if the key doesn't exist in cache.
        :return: Tuple with last_write time, and the stored object.
        """

        cache_contents = self.__read()
        contents: Dict = jsonpickle.decode(cache_contents)
        cache = contents.get(cache_key, {})
        result = cache.get(self._LAST_WRITE_KEY, 0), cache.get(self._STORE_KEY, default)
        log.info(f'Returning items from cache for cache key: {cache_key}')
        return result

    def get_val(self, cache_key: str, default=None) -> Any:
        return self.get(cache_key, default)[1]

    @prime_cache
    @wipe_bad_cache
    def append(self, cache_key: str, objects: Union[Dict, Set[Any]]):
        """
        Add a set or dictionary of items to the existing cache. Must be the SAME type as what is stored in the
        existing cache under cache_key.

        Lists not supported due to issues with duplicates continually growing in the local cache :)

        :param cache_key: Key to append or merge items with
        :param objects: Objects to add.
        """

        if isinstance(objects, Set):
            objects = list(objects)

        if len(objects) > 0:
            log.info(f'Appending {len(objects)} items to local cache: {objects}')

            contents: Dict = jsonpickle.decode(self.__read())
            cache = contents.get(cache_key, {})
            refresh_time = cache.get(self._LAST_REFRESH_KEY, 0)
            cache_obj = cache.get(self._STORE_KEY)

            if isinstance(cache_obj, Dict) and isinstance(objects, Dict) or cache_obj is None:
                if cache_obj:
                    cache_obj.update(objects)
                else:
                    cache_obj = objects
            elif isinstance(cache_obj, List) and isinstance(objects, List) or cache_obj is None:
                cache_obj = list(set(cache_obj + objects)) if cache_obj else objects
            else:
                raise RuntimeError(
                    "Invalid state detected. Cache contains an invalid type that cannot be appended to, "
                    "or the type provided does not match the type stored in the cache.")

            contents[cache_key] = {
                self._STORE_KEY: cache_obj,
                self._LAST_WRITE_KEY: Utils.millis_since_epoch(),
                self._LAST_REFRESH_KEY: refresh_time,
            }

            self.__write(jsonpickle.dumps(contents))
        else:
            log.info('No cached items found to add to cache.')

    def delete(self, cache_key: str, objects: Union[Dict, Set[Any]]):
        """
        If any of these items exist in the cache for this set of stored values, delete them.
        :param cache_key: Key to potentially delete items from
        :param objects: *Keys* in the cached DICT to delete, or items in a cached LIST to delete.
        """

        if isinstance(objects, Set):
            objects = list(objects)

        if len(objects) > 0:
            log.info(f'Deleting {len(objects)} items from local cache: {objects}')

            contents: Dict = jsonpickle.decode(self.__read())
            cache = contents.get(cache_key, {})
            refresh_time = cache.get(self._LAST_REFRESH_KEY, 0)
            cache_obj = cache.get(self._STORE_KEY)
            log.info(f'In cache: {cache_obj}')

            if isinstance(cache_obj, Dict) and isinstance(objects, Dict) or cache_obj is None:
                log.info(f'Cache Obj is a dict')
                if cache_obj:
                    for obj in objects:
                        del cache_obj[obj]

            elif isinstance(cache_obj, List) and isinstance(objects, List) or cache_obj is None:
                log.info(f"Cache obj is a list..")
                if cache_obj:
                    cache_obj = list(set(cache_obj) - set(objects))

            else:
                raise RuntimeError(
                    "Invalid state detected. Cache contains an invalid type that cannot be appended to, "
                    "or the type provided does not match the type stored in the cache.")

            log.info(f'New cache obj: {cache_obj}')
            contents[cache_key] = {
                self._STORE_KEY: cache_obj,
                self._LAST_WRITE_KEY: Utils.millis_since_epoch(),
                self._LAST_REFRESH_KEY: refresh_time,
            }

            self.__write(jsonpickle.dumps(contents))
        else:
            log.info('No cached items found to add to cache.')

    def wipe_cache(self):
        self.__write(jsonpickle.dumps({}))



import sys

import pexpect

from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.utils.utils import *


class DevCleanup(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)

    def run(self):
        self.cleanup_success()
        self.cleanup_with_orphans()

    def cleanup_success(self):
        self.step(f"Testing: {CLI_NAME} config {Utils.get_first(cleanup)} --env {DEFAULT_ENV} "
              f"--config figcli/test/assets/success/figgy.json --skip-upgrade ")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(cleanup)} --env {DEFAULT_ENV} '
                                    f'--config figcli/test/assets/success/figgy.json --skip-upgrade {self.extra_args}',
                                    encoding='utf-8', timeout=10)
        child.expect('.*No orphaned keys.*No remote replication configs.*')
        print("Empty cleanup, success!")

    def cleanup_with_orphans(self):
        self.step(f"Testing: {CLI_NAME} config {Utils.get_first(cleanup)} --env {DEFAULT_ENV} "
              f"--config figcli/test/assets/error/figgy.json")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(cleanup)} --env {DEFAULT_ENV} '
                                    f'--config figcli/test/assets/error/figgy.json --skip-upgrade {self.extra_args}',
                                    encoding='utf-8', timeout=10)
        child.expect('.*/app/ci-test/v1/config11.* exists.*but does not exist.*')
        child.sendline('n')
        child.expect('.*replication mapping.*does not exist.*')
        child.sendline('n')
        print("Cleanup with orphans success!")


from dataclasses import dataclass
from typing import Optional, Tuple, Dict
import requests
import logging
import random
import re
from figcli.config.constants import *
from figcli.config.style.color import Color
from figcli.config.style.terminal_factory import TerminalFactory
from figcli.models.defaults.defaults import CLIDefaults

log = logging.getLogger(__name__)


@dataclass
class FiggyVersionDetails:
    version: str
    notify_chance: int
    changelog: str

    def changes_from(self, old_version: str):
        regex = f'.*(##+\s+{self.version}.*)##+\s+{old_version}.*'
        result = re.match(regex, self.changelog, re.DOTALL)
        if result:
            return result.group(1).rstrip()
        else:
            return f"Unable to parse changes for new version: {self.version}"

    @staticmethod
    def from_api_response(response: Dict) -> "FiggyVersionDetails":
        notify_chance, version, changelog = \
            response.get('notify_chance'), response.get('version'), response.get('changelog')

        if not notify_chance or notify_chance == 'None':
            notify_chance = 0

        if not version:
            raise ValueError('No valid version found.')
        elif not changelog:
            raise ValueError('No valid changelog found.')

        return FiggyVersionDetails(
            version=version,
            notify_chance=int(notify_chance),
            changelog=response.get('changelog')
        )


class VersionTracker:
    _UPGRADE_CHECK_PERCENTAGE = 5  # % chance any decorated method execution will check for an upgrade

    def __init__(self, cli_defaults: CLIDefaults):
        self._cli_defaults = cli_defaults
        self.c = TerminalFactory(self._cli_defaults.colors_enabled).instance().get_colors()

    @staticmethod
    def get_version() -> FiggyVersionDetails:
        result = requests.get(FIGGY_GET_VERSION_URL)
        if result.status_code == 200:
            details: FiggyVersionDetails = FiggyVersionDetails.from_api_response(result.json())
            return details
        else:
            raise ValueError("Unable to fetch figgy version details.")

    @staticmethod
    def check_version(c: Color) -> None:
        try:
            new_details = VersionTracker.get_version()

            if new_details.version != VERSION:
                VersionTracker.print_new_version_msg(c, new_details)
                VersionTracker.print_changes(c, new_details)
            else:
                print(f"Version: {VERSION}.")
                print(f"You are currently running the latest version of figgy.")

        except ValueError as e:
            log.warning("Unable to fetch version information from remote endpoint.")
            print(f"Version: {VERSION}")

    @staticmethod
    def print_changes(c: Color, new_details: FiggyVersionDetails) -> None:
        if new_details.version != VERSION:
            print(f'\n\n{c.fg_yl}Changes you\'ll get if you upgrade!{c.rs}')
            print(f'{c.fg_bl}------------------------------------------{c.rs}')
            print(new_details.changes_from(VERSION))
            print(f'{c.fg_bl}------------------------------------------{c.rs}')

    @staticmethod
    def print_new_version_msg(c: Color, new_details: FiggyVersionDetails):
        if not VersionTracker.is_rollback(VERSION, new_details.version):
            print(f'\n{c.fg_bl}------------------------------------------{c.rs}')
            print(f'A new version of figgy is available!')
            print(f"Current Version: {c.fg_yl}{VERSION}{c.rs}")
            print(f"New Version: {c.fg_bl}{new_details.version}{c.rs}")
            print(f"To see what the new version has in store for you, run `{CLI_NAME} --version`")
            print(f"To upgrade, run `brew upgrade figgy` or `pip install figgy-cli --upgrade`")
            print(f'{c.fg_bl}------------------------------------------{c.rs}')
        else:
            print(f'\n{c.fg_bl}------------------------------------------{c.rs}')
            print(f'Figgy was rolled back due to a major issue and you\'re on a bad version!')
            print(f"Current Version: {c.fg_yl}{VERSION}{c.rs}")
            print(f"Recommended Version: {c.fg_bl}{new_details.version}{c.rs}")
            print(f"To roll-back, run `brew upgrade figgy` or `pip install figgy-cli --upgrade`")
            print(f'{c.fg_bl}------------------------------------------{c.rs}')

    @staticmethod
    def is_rollback(current_version: str, new_version: str):
        try:
            cu_major = current_version.split('.')[0]
            cu_minor = current_version.split('.')[1]
            cu_patch = current_version.split('.')[2].strip('ab')
            new_major = new_version.split('.')[0]
            new_minor = new_version.split('.')[1]
            new_patch = new_version.split('.')[2].strip('ab')

            if new_major < cu_major:
                return True
            elif new_major <= cu_major and new_minor < cu_minor:
                return True
            elif new_major <= cu_major and new_minor <= cu_minor and new_patch < cu_patch:
                return True
            else:
                return False
        except IndentationError:
            pass

    @staticmethod
    def upgrade_available(current_version: str, new_version: str):
        try:
            cu_major = current_version.split('.')[0]
            cu_minor = current_version.split('.')[1]
            cu_patch = current_version.split('.')[2].strip('ab')
            new_major = new_version.split('.')[0]
            new_minor = new_version.split('.')[1]
            new_patch = new_version.split('.')[2].strip('ab')

            if new_major > cu_major:
                return True
            elif new_major >= cu_major and new_minor > cu_minor:
                return True
            elif new_major >= cu_major and new_minor >= cu_minor and new_patch > cu_patch:
                return True
            else:
                return False
        except IndentationError:
            pass

    @staticmethod
    def notify_user(function):
        """
        Has a _chance_ to notify a user if a new version has been released and they are on the old version.
        """

        log.info("Rolling dice for update notify chance")

        def inner(self, *args, **kwargs):
            log.info("Rolling dice to check version..")
            if VersionTracker._UPGRADE_CHECK_PERCENTAGE > random.randint(0, 99):
                log.info("Checking for new version..")
                try:
                    details = VersionTracker.get_version()
                    if details.notify_chance > random.randint(0, 99) and details.version != VERSION:
                        log.info("Notifying user of new version")
                        if hasattr(self, 'c') and isinstance(self.c, Color):
                            VersionTracker.print_new_version_msg(self.c, details)
                except ValueError:
                    log.warning("Unable to fetch version information from remote resource.")

            return function(self, *args, **kwargs)

        return inner



from prompt_toolkit.completion import WordCompleter

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *


class Audit(ConfigCommand):
    """
    Returns audit history for a queried PS Name
    """
    def __init__(self, ssm_init: SsmDao, config_init: ConfigDao, config_completer_init: WordCompleter,
                 colors_enabled: bool, config_context: ConfigContext):
        super().__init__(audit, colors_enabled, config_context)
        self._ssm = ssm_init
        self._config = config_init
        self._config_completer = config_completer_init
        self._utils = Utils(colors_enabled)

    def _audit(self):
        audit_more = True

        while audit_more:
            ps_name = Input.input(f"Please input a PS Name : ", completer=self._config_completer)
            audit_logs = self._config.get_audit_logs(ps_name)
            result_count = len(audit_logs)
            if result_count > 0:
                print(f"\nFound {self.c.fg_bl}{result_count}{self.c.rs} results.")
            else:
                print(f"\n{self.c.fg_yl}No results found for: {ps_name}{self.c.rs}")
            for log in audit_logs:
                print(log)

            to_continue = input(f"Audit another? (Y/n): ")
            to_continue = to_continue if to_continue != '' else 'y'
            audit_more = to_continue.lower() == "y"
            print()

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._audit()



import re
from typing import *

from botocore.exceptions import ClientError

from figcli.commands.config.get import Get
from figcli.commands.config.put import Put
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.extras.key_utils import KeyUtils
from figcli.models.replication_config import ReplicationConfig, ReplicationType
from figcli.config import *
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import Utils


class Sync(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, config_init: ConfigDao, colors_enabled: bool,
                 context: ConfigContext, get: Get, put: Put):
        super().__init__(sync, colors_enabled, context)
        self._config = config_init
        self._ssm = ssm_init
        self._config_path = context.ci_config_path if context.ci_config_path else Utils.find_figgy_json()
        self._utils = Utils(colors_enabled)
        self._replication_only = context.replication_only
        self._errors_detected = False
        self.example = f"{self.c.fg_bl}{CLI_NAME} config {self.command_printable} " \
                       f"--env dev --config /path/to/config{self.c.rs}"
        self._get: Get = get
        self._put: Put = put
        self._FILE_PREFIX = "file://"

    def _load_file(self, file_path: str) -> str:
        try:
            with open(file_path, 'r') as file:
                return file.read()
        except FileNotFoundError:
            self._utils.error_exit(f"Provided file path: {file_path} is invalid. No file found.")

    def _input_config_values(self, config_keys: Set[str]) -> None:
        """
        Prompts the user for each of the passed in set of config values if any are missing from PS.
        :param config_keys: Set[string] - config values to prompt the user to add.
        """

        def validate_msg(ps_name: str):
            print(f"Name Validated: {self.c.fg_bl}{ps_name}{self.c.rs}")
            return validate_msg

        count = 0
        for key in config_keys:
            try:
                if not self._get.get(key):
                    print(f"Name {self.c.fg_bl}{key}{self.c.rs} does not exist in ParameterStore for the "
                          f"{self.c.fg_bl}{self.run_env}{self.c.rs} environment. It must be added before this "
                          f"build can continue.\n")
                    self._put.put_param(key=key, display_hints=count == 0)
                    count = count + 1
                else:
                    validate_msg(key)
            except ClientError:
                validate_msg(key)

        if count:
            print(f"{self.c.fg_bl}{count} {'value' if count == 1 else 'values'} added successfully{self.c.rs}")

    def _sync_keys(self, config_namespace: str, all_keys: Set):
        """
        Looks for orphaned parameters (keys) under the namespace provided and prints out information about
        missing parameters that are not defined in the figgy.json file
        Args:
            config_namespace: Namespace to query PS under.
            all_keys: All keys that exist in figgy.json to compare against.
        """
        print(f"{self.c.fg_gr}Checking for orphaned config names.{self.c.rs}")

        # Find & Cleanup orphaned keys
        ps_keys = set(list(map(lambda x: x['Name'], self._ssm.get_all_parameters([config_namespace]))))
        ps_only_keys = ps_keys.difference(all_keys)

        if len(ps_only_keys) > 0:
            print(UNUSED_CONFIG_DETECTED.replace("%%red%%", self.c.fg_rd)
                  .replace("%%blue%%", self.c.fg_bl)
                  .replace("%%rs%%", self.c.rs))

        for key in ps_only_keys:
            print(f"Unused Parameter: {self.c.fg_bl}{key}{self.c.rs}")

        if not ps_only_keys:
            print(f"{self.c.fg_bl}No orphaned names found.{self.c.rs}")

    def _sync_repl_configs(self, config_repl: Dict, namespace: str = None) -> None:
        """
        Syncs replication configs from a defined "replicate_figs" block parsed from either the figgy.json file
        or the data replication config json file.
        Args:
            config_repl: Dict of KV Pairs for a repl config. Source -> Dest
            namespace: Optional namespace. Parsed from destination if not supplied.
        """
        local_configs: List[ReplicationConfig] = ReplicationConfig.from_dict(conf=config_repl,
                                                                             type=ReplicationType(REPL_TYPE_APP),
                                                                             run_env=self.run_env, namespace=namespace)
        for l_cfg in local_configs:
            # Namespace will be missing for --replication-only syncs. Otherwise, with standard syncs, namespace is passed
            # as a parameter here.
            if not namespace:
                namespace = l_cfg.namespace

            if not l_cfg.destination.startswith(namespace):
                print(f"{self.c.fg_rd}Replication config {l_cfg.source} -> {l_cfg.destination} has a destination that "
                      f"is not in your service namespace: {namespace}. This is invalid.{self.c.rs}")
                self.errors_detected = True
                continue

            remote_cfg = self._config.get_config_repl(l_cfg.destination, self.run_env)

            # Should never happen, except when someone manually deletes source / destination without going through CLI
            missing_from_ps = self._ssm.get_parameter_encrypted(l_cfg.source) is None

            if not remote_cfg or remote_cfg != l_cfg or missing_from_ps:
                try:
                    if self._can_replicate_from(l_cfg.source) and not remote_cfg or missing_from_ps:
                        self._config.put_config_repl(l_cfg)
                        print(f"{self.c.fg_gr}Replication added:{self.c.rs} {l_cfg.source} -> {l_cfg.destination}")
                    elif self._can_replicate_from(l_cfg.source) and remote_cfg:
                        self._config.put_config_repl(l_cfg)
                        print(f"{self.c.fg_bl}Replication updated.{self.c.rs}")
                        print(f"{self.c.fg_rd}Removed: {remote_cfg.source} -> {remote_cfg.destination}")
                        print(f"{self.c.fg_gr}Added: {l_cfg.source} -> {l_cfg.destination}")
                    else:
                        self._errors_detected = True
                        # print(f"{self.c.fg_rd}You do not have permission to configure replication from source:"
                        #       f"{self.c.rs} {key}")
                except ClientError:
                    self._utils.validate(False, f"Error detected when attempting to store replication config "
                                                f"for {l_cfg.destination}")
                    self._errors_detected = True
            else:
                print(f"{self.c.fg_bl}Replication Validated:{self.c.rs} {l_cfg.source} -> {l_cfg.destination}")

    def _notify_of_data_repl_orphans(self, config_repl: Dict) -> None:
        """
        Notify user of detected orphaned replication configurations when using the --replication-only flag.
        :param config_repl: replication configuration block.
        """
        orphans: Set[ReplicationConfig] = set()
        notify = False
        for repl in config_repl:
            namespace = self._utils.parse_namespace(config_repl[repl])
            remote_cfgs = self._config.get_all_configs(self.run_env, namespace)

            if remote_cfgs:
                for cfg in remote_cfgs:
                    if cfg.source not in list(config_repl.keys()) \
                            and cfg.type == REPL_TYPE_APP \
                            and not cfg.source.startswith(shared_ns) \
                            and not cfg.source.startswith(self.context.defaults.service_ns):
                        orphans.add(cfg)
                        notify = True

        for orphan in orphans:
            print(f"{self.c.fg_rd}Orphaned replication mapping detected: {self.c.rs}"
                  f" {self.c.fg_bl}{orphan.source} -> {orphan.destination}{self.c.rs}.")
        if notify:
            print(f"To cleanup the now orphaned replication configs, "
                  f"delete the destination, THEN the source with the `figgy config delete` command")

    def _sync_replication(self, config_repl: Dict, expected_destinations: Set, namespace: str):
        """
        Calls sync_repl_configs which adds/removes repl configs. Then searches for orphaned configurations and notifies
        the user of detected orphaned configurations.
        Args:
            config_repl: Dict of KV Pairs for a repl config. Source -> Dest
            expected_destinations: expected replication destinations, as defined in merge key sources,
             or shared_figs
            namespace: Namespace to sync replication configs to. E.g. /app/demo-time/
        """

        print(f"{self.c.fg_gr}Validating replication for all parameters.{self.c.rs}")

        self._sync_repl_configs(config_repl, namespace=namespace)
        print(f"\n{self.c.fg_gr}Checking for orphaned replication configurations.{self.c.rs}")
        remote_cfgs = self._config.get_all_configs(self.run_env, namespace)
        notify = True
        if remote_cfgs:
            for cfg in remote_cfgs:
                if cfg.source not in list(config_repl.keys()) \
                        and cfg.destination not in list(config_repl.values()) \
                        and cfg.destination not in expected_destinations \
                        and (isinstance(cfg.source, list)
                             or cfg.source.startswith(shared_ns) or cfg.source.startswith(
                            self.context.defaults.service_ns)):
                    print(f"{self.c.fg_rd}Orphaned replication mapping detected: {self.c.rs}"
                          f" {self.c.fg_bl}{cfg.source} -> {cfg.destination}{self.c.rs}.")
                    notify = False
        if notify:
            print(f"{self.c.fg_bl}No orphaned replication configs found for: "
                  f"{self.c.rs}{self.c.fg_gr}{namespace}{self.c.rs}")
        else:
            print(f"{self.c.fg_bl}{CLEANUP_REPLICA_ORPHANS}{self.c.rs}")

    def _validate_merge_keys(self, destination: str, sources: Union[List, str], namespace: str) -> bool:
        """
        Validates merge key sources & destinations
        Args:
            destination: str -> Destination of merge key replication
            sources: List or Str -> Source(e) of this merge key
            namespace: application namespace
        """
        if not destination.startswith(namespace):
            print(f"{self.c.fg_rd}Merge config: {self.c.rs}{self.c.fg_bl}{destination}{self.c.rs}{self.c.fg_rd} has a "
                  f"destination that is not in your service namespace: "
                  f"{self.c.rs}{self.c.fg_bl}{namespace}{self.c.rs}{self.c.fg_rd}. This is invalid.{self.c.rs}")
            self.errors_detected = True
            return False

        if isinstance(sources, list):
            for item in sources:
                if item.startswith(MERGE_KEY_PREFIX):
                    self._utils.validate(item.replace(MERGE_KEY_PREFIX, "").startswith(namespace),
                                         f"Source: {item} in merge config must begin with your namespace: {namespace}.")
                    self.errors_detected = True
                    return False
        else:
            self._utils.validate(sources.startswith(namespace),
                                 f"Source {sources} in merge config must begin with your namespace: {namespace}")
            self.errors_detected = True
            return False

        return True

    def _sync_merge_keys(self, config_merge: Dict, namespace: str) -> None:
        """
            Pushes merge key configs into replication config table.
        Args:
            config_merge: Dict of merge_parameters parsed from figcli.json file
            namespace: namespace for app
        """
        print(f"{self.c.fg_gr}Validating replication for all merge keys.{self.c.rs}")
        for key in config_merge:
            self._validate_merge_keys(key, config_merge[key], namespace)

            config = self._config.get_config_repl(key, self.run_env)
            if not config or (config.source != config_merge[key]):
                try:
                    repl_config = ReplicationConfig(key, self.run_env, namespace,
                                                    config_merge[key], ReplicationType(REPL_TYPE_MERGE))
                    self._config.put_config_repl(repl_config)
                except ClientError:
                    self._utils.validate(False, f"Error detected when attempting to store replication config for {key}")
                    self._errors_detected = True
            else:
                print(f"Merge key replication config validated: {key}")

    def _validate_expected_names(self, all_names: Set, repl_conf: Dict, merge_conf: Dict):
        print(f"{self.c.fg_gr}Validating shared keys exist.{self.c.rs}")
        print_resolution_message = False
        merged_confs = {**repl_conf, **merge_conf}
        for name in all_names:
            if self._ssm.get_parameter_encrypted(name) is None:
                awaiting_repl = False
                for cnf in merged_confs:
                    if name == cnf or name in list(repl_conf.values()):
                        print(f"{self.c.fg_bl}Config value {name} is a destination for replication, but doesn't exist "
                              f"yet. If you commit now your build could fail. This will auto-resolve itself if all of "
                              f"its dependencies exist.{self.c.rs}")
                        awaiting_repl = True
                        break

                if not awaiting_repl:
                    print(f"Config value of {self.c.fg_rd}{name}{self.c.rs} does not exist and is expected based on"
                          f" your defined configuration.")
                    print_resolution_message = True
                    self._errors_detected = True

        if print_resolution_message:
            print(f"\n\n{self.c.fg_rd}{SHARED_NAME_RESOLUTION_MESSAGE}{self.c.rs}")

    def _can_replicate_from(self, source: str):
        try:
            if self._ssm.get_parameter_encrypted(source) is not None:
                return True
            else:
                print(f"{self.c.fg_yl}Replication source: {source} is missing from ParameterStore. "
                      f"It must be added before config replication can be configured.{self.c.rs}\n")
                self._input_config_values({source})
                return True
        except ClientError as e:
            denied = "AccessDeniedException" == e.response['Error']['Code']
            if denied and "AWSKMS; Status Code: 400;" in e.response['Error']['Message']:
                print(f"{self.c.fg_rd}You do not have access to decrypt the value of Name: {source}{self.c.rs}")
            elif denied:
                print(f"{self.c.fg_rd}You do not have access to Parameter: {source}{self.c.rs}")
            else:
                raise
        return False

    def _validate_replication_config(self, config_repl: Dict, app_conf: bool = True):
        """
        Validates replication config blocks are valid / legal. Prevents people from setting up replication from
        disallowed namespaces, etc. Exits with error if invalid config is discovered.

        Args:
            config_repl: Dict of KV Pairs for a repl config. Source -> Dest
            app_conf: bool: T/F - True if this is an application config block in an application config (figgy.json).
                    False if other, which for now is only repl-configs for data teams.
        """
        for key in config_repl:
            if app_conf:
                self._utils.validate(re.match(f'^/shared/.*$|^{self.context.defaults.service_ns}/.*$', key) is not None,
                                     f"The SOURCE of your replication configs must begin with `/shared/` or "
                                     f"`{self.context.defaults.service_ns}/`. "
                                     f"{key} is non compliant.")

            self._utils.validate(re.match(f'^{self.context.defaults.service_ns}/.*$', config_repl[key]) is not None,
                                 f"The DESTINATION of your replication configs must always begin with "
                                 f"`{self.context.defaults.service_ns}/`")

    def _find_missing_shared_figs(self, namespace: str, config_repl: Dict, shared_names: set, merge_conf: Dict):
        """
            Notifies the user if there is a parameter that has been shared into their namespace by an outside party
            but they have not added it to the `shared_figs` block of their figgy.json
        """
        all_repl_cfgs = self._config.get_all_configs(self.run_env, namespace)
        for cfg in all_repl_cfgs:
            in_merge_conf = self._in_merge_value(cfg.destination, merge_conf)

            if cfg.destination not in shared_names and cfg.type == REPL_TYPE_APP \
                    and cfg.destination not in config_repl.values() and not in_merge_conf:
                print(f"It appears that {self.c.fg_bl}{cfg.user}{self.c.rs} shared "
                      f"{self.c.fg_bl}{cfg.source}{self.c.rs} to {self.c.fg_bl}{cfg.destination}{self.c.rs} "
                      f"and you have not added {self.c.fg_bl}{cfg.destination}{self.c.rs} to the "
                      f"{self.c.fg_bl}{SHARED_KEY}{self.c.rs} section of your figgy.json. This is also not "
                      f"referenced in any defined merge parameter. Please add "
                      f"{self.c.fg_bl}{cfg.destination}{self.c.rs} to your figgy.json, or delete this parameter "
                      f"and the replication config with the cleanup command.")

    def _in_merge_value(self, dest: str, merge_conf: Dict):
        for key in merge_conf:
            value = merge_conf[key]
            # 'value' can be a list or a str, but the way 'in' operates, this works either way. #dynamic programming
            for suffix in merge_suffixes:
                if f"${'{'}{dest}{suffix}{'}'}" in value:
                    return True

        return False

    def run_ci_sync(self) -> None:
        """
            Orchestrates a standard `sync` command WITHOUT The `--replication-only` flag set.
        """
        # Validate & parse figgy.json
        config = self._utils.get_ci_config(self._config_path)
        shared_names = set(self._utils.get_config_key_safe(SHARED_KEY, config, default=[]))
        repl_conf = self._utils.get_config_key_safe(REPLICATION_KEY, config, default={})
        repl_from_conf = self._utils.get_config_key_safe(REPL_FROM_KEY, config, default={})
        merge_conf = self._utils.get_config_key_safe(MERGE_KEY, config, default={})
        config_keys = set(self._utils.get_config_key_safe(CONFIG_KEY, config, default=[]))
        namespace = self._utils.get_namespace(config)
        merge_keys = set(merge_conf.keys())
        all_keys = KeyUtils.find_all_expected_names(config_keys, shared_names, merge_conf, repl_conf,
                                                    repl_from_conf, namespace)

        repl_conf = KeyUtils.merge_repl_and_repl_from_blocks(repl_conf, repl_from_conf, namespace)
        # Add missing config values
        print()
        print(f"{self.c.fg_gr}Validating all configuration keys exist in ParameterStore.{self.c.rs}")
        self._input_config_values(config_keys)

        # Sync keys between PS / Local config
        print()
        self._sync_keys(namespace, all_keys)

        print()

        self._find_missing_shared_figs(namespace, repl_conf, shared_names, merge_conf)

        # Disabling requirement (for now) of replication to be in /replicated path
        # print()
        self._validate_replication_config(repl_conf, app_conf=True)

        print()
        # sync replication config
        all_shared_keys = shared_names | set(merge_conf.keys())
        self._sync_replication(repl_conf, all_shared_keys, namespace)

        print()
        self._sync_merge_keys(merge_conf, namespace)

        print()
        # validate expected keys exist
        self._validate_expected_names(all_keys, repl_conf, merge_conf)

    def run_repl_sync(self) -> None:
        """
        Orchestrates sync when the user passes in the `--replication-only` flag.
        """
        repl_conf = self._utils.get_repl_config(self._config_path)
        self._validate_replication_config(repl_conf, app_conf=False)
        self._sync_repl_configs(repl_conf)
        self._notify_of_data_repl_orphans(repl_conf)

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        print()
        if self._replication_only:
            self.run_repl_sync()
        else:
            self.run_ci_sync()

        if self._errors_detected:
            print(f"\r\n{self.c.fg_rd}Sync failed. Please address the outputted errors.{self.c.rs}")
        else:
            print(f"\r\n{self.c.fg_gr}Sync completed with no errors!{self.c.rs}")



import os
import readline
import stat
import sys
from urllib.request import urlopen

import requests
from zipfile import ZipFile

import tqdm
from figcli.extras.completer import Completer

from figcli.config import HOME, CLI_NAME
from figcli.svcs.observability.version_tracker import FiggyVersionDetails, VersionTracker
from figcli.utils.utils import Utils

class UpgradeManager:
    def __init__(self, colors_enabled: bool):
        self._utils = Utils(colors_enabled=colors_enabled)
        self.c = Utils.default_colors(enabled=colors_enabled)
        self.current_version: FiggyVersionDetails = VersionTracker.get_version()

    def download_zip(self, remote_path: str, local_path: str):
        eg_link = remote_path
        response = requests.get(eg_link, stream=True)
        with tqdm.wrapattr(open(local_path, "wb"), "write",
                           miniters=1, desc=eg_link.split('/')[-1],
                           total=int(response.headers.get('content-length', 0))) as fout:
            for chunk in response.iter_content(chunk_size=4096):
                fout.write(chunk)

    def download_from_url(self, url, dst):
        """
        @param: url to download file
        @param: dst place to put the file
        """
        file_size = int(urlopen(url).info().get('Content-Length', -1))
        if os.path.exists(dst):
            first_byte = os.path.getsize(dst)
        else:
            first_byte = 0
        if first_byte >= file_size:
            return file_size
        header = {"Range": "bytes=%s-%s" % (first_byte, file_size)}
        pbar = tqdm(
            total=file_size, initial=first_byte,
            unit='B', unit_scale=True, desc=url.split('/')[-1])
        req = requests.get(url, headers=header, stream=True)
        with(open(dst, 'ab')) as f:
            for chunk in req.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    pbar.update(1024)
        pbar.close()
        return file_size

    def install_mac_onedir(self, install_path: str, latest_version: str):
        zip_path = f"{HOME}/.figgy/figgy.zip"
        install_dir = f'{HOME}/.figgy/installations/{latest_version}'
        remote_path = f'http://www.figgy.dev/releases/cli/{latest_version}/darwin/figgy.zip'
        os.makedirs(os.path.dirname(install_dir), exist_ok=True)
        self.download_from_url(remote_path, zip_path)

        with ZipFile(zip_path, 'r') as zipObj:
            zipObj.extractall(install_dir)

        if self._utils.file_exists(install_path):
            os.remove(install_path)

        executable_path = f'{install_dir}/{CLI_NAME}'
        st = os.stat(executable_path)
        os.chmod(executable_path, st.st_mode | stat.S_IEXEC)
        os.symlink(f'{install_dir}/{CLI_NAME}', install_path)
        print(f'{CLI_NAME} has been installed at path `{install_path}`.')

    def get_install_path(self) -> str:
        """
        Prompts the user to get their local installation path.
        """
        readline.parse_and_bind("tab: complete")
        comp = Completer()
        readline.set_completer_delims(' \t\n')
        readline.parse_and_bind("tab: complete")
        readline.set_completer(comp.pathCompleter)
        abs_path = os.path.dirname(sys.executable)
        install_path = input(f'Input path to your existing installation. Default: {abs_path} : ') or abs_path
        suffix = ".exe" if self._utils.is_windows() else ""

        if os.path.isdir(install_path):
            if install_path.endswith('/'):
                install_path = install_path[:-1]

            install_path = f"{install_path}/{CLI_NAME}{suffix}"

        if not self._utils.file_exists(install_path):
            self._utils.error_exit("Invalid install path specified, try providing the full path to the binary.")

        return install_path



from figcli.data.dao.ssm import SsmDao
from botocore.exceptions import ClientError
import botocore.session
from typing import Dict
from figcli.utils.utils import Utils


class SSMRebirth:
    def __init__(self):
        aws_session = botocore.session.get_session()
        boto_client = aws_session.create_client("ssm", "us-east-1")
        self._ssm: SsmDao = SsmDao(boto_client)
        self._utils = Utils(True)

    def _client_exception_msg(self, item: Dict, e: ClientError):
        if "AccessDeniedException" == e.response["Error"]["Code"]:
            print(
                f"\n\nYou do not have permissions to a new config value at the path: "
                f"{item['parameter_name']}"
            )
        else:
            print(f"Error message: {e.response['Error']['Message']}")

    def start(self):
        items = self._ssm.get_all_parameters("/")

        for item in items:

            try:
                key_id = None if item["Type"] == "String" else item["KeyId"]

                # Grab decrypted value from SSM for key
                value = self._ssm.get_parameter(item["Name"])
                print(f"Key {item['Name']} has value: {value}")

                description = item["Description"] if "Description" in item else ""

                # Test to make sure we can write to this parameter path
                # If this fails, an exception should be thrown.
                self._ssm.set_parameter(
                    item["Name"],
                    value,
                    description,
                    item["Type"],
                    key_id,
                )

                print(f"Deleting: {item['Name']}")
                self._ssm.delete_parameter(item["Name"])

                print(f"Recreating: {item['Name']}")
                self._ssm.set_parameter(
                    item["Name"],
                    value,
                    description,
                    item["Type"],
                    key_id,
                )

            except ClientError as e:
                self._client_exception_msg(item, e)


if __name__ == "__main__":
    rebirth = SSMRebirth()
    rebirth.start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import base64
import io
import json
import logging
import os
import re
import sys
import requests
import subprocess

from distutils.spawn import find_executable
from bs4 import BeautifulSoup
from requests import HTTPError
from six import print_ as print
from six.moves import urllib_parse, input
from figcli.svcs.mfa.u2f import *

class ExpectedGoogleException(Exception):
    def __init__(self, *args):
        super(ExpectedGoogleException, self).__init__(*args)


class Google:
    def __init__(self, config, save_failure):
        """The Google object holds authentication state
        for a given session. You need to supply:

        username: FQDN Google username, eg first.last@example.com
        password: obvious
        idp_id: Google's assigned IdP identifier for your G-suite account
        sp_id: Google's assigned SP identifier for your AWS SAML app

        Optionally, you can supply:
        duration_seconds: number of seconds for the session to be active (max 43200)
        """

        # self.version = _version.__version__
        self.version = "0.0.35"
        self.config = config
        self.base_url = 'https://accounts.google.com'
        self.save_failure = save_failure
        self.session_state = None

    @property
    def login_url(self):
        return self.base_url + "/o/saml2/initsso?idpid={}&spid={}&forceauthn=false".format(
            self.config.idp_id, self.config.sp_id)

    def check_for_failure(self, sess):

        if isinstance(sess.reason, bytes):
            # We attempt to decode utf-8 first because some servers
            # choose to localize their reason strings. If the string
            # isn't utf-8, we fall back to iso-8859-1 for all other
            # encodings. (See PR #3538)
            try:
                reason = sess.reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = sess.reason.decode('iso-8859-1')
        else:
            reason = sess.reason

        if sess.status_code == 403:
            raise ExpectedGoogleException(u'{} accessing {}'.format(
                reason, sess.url))

        try:
            sess.raise_for_status()
        except HTTPError as ex:

            if self.save_failure:
                logging.exception("Saving failure trace in 'failure.html'", ex)
                with open("failure.html", 'w') as out:
                    out.write(sess.text)

            raise ex

        return sess

    def post(self, url, data=None, json=None):
        try:
            response = self.check_for_failure(self.session.post(url, data=data, json=json))
        except requests.exceptions.ConnectionError as e:
            logging.exception(
                'There was a connection error, check your network settings.', e)
            sys.exit(1)
        except requests.exceptions.Timeout as e:
            logging.exception('The connection timed out, please try again.', e)
            sys.exit(1)
        except requests.exceptions.TooManyRedirects as e:
            logging.exception('The number of redirects exceeded the maximum '
                              'allowed.', e)
            sys.exit(1)

        return response

    def get(self, url):
        try:
            response = self.check_for_failure(self.session.get(url))
        except requests.exceptions.ConnectionError as e:
            logging.exception(
                'There was a connection error, check your network settings.', e)
            sys.exit(1)
        except requests.exceptions.Timeout as e:
            logging.exception('The connection timed out, please try again.', e)
            sys.exit(1)
        except requests.exceptions.TooManyRedirects as e:
            logging.exception('The number of redirects exceeded the maximum '
                              'allowed.', e)
            sys.exit(1)

        return response

    @staticmethod
    def parse_error_message(sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        error = response_page.find('span', {'id': 'errorMsg'})

        if error is None:
            return None
        else:
            return error.text

    @staticmethod
    def find_key_handles(input, challengeTxt):
        keyHandles = []
        typeOfInput = type(input)
        if typeOfInput == dict:  # parse down a dict
            for item in input:
                keyHandles.extend(Google.find_key_handles(input[item], challengeTxt))

        elif typeOfInput == list:  # looks like we've hit an array - iterate it
            array = list(filter(None, input))  # remove any None type objects from the array
            for item in array:
                typeValue = type(item)
                if typeValue == list:  # another array - recursive call
                    keyHandles.extend(Google.find_key_handles(item, challengeTxt))
                elif typeValue == int or typeValue == bool:  # ints bools etc we don't care
                    continue
                else:  # we went a string or unicode here (python 3.x lost unicode global)
                    try:  # keyHandle string will be base64 encoded -
                        # if its not an exception is thrown and we continue as its not the string we're after
                        base64UrlEncoded = base64.urlsafe_b64encode(base64.b64decode(item))
                        if base64UrlEncoded != challengeTxt:  # make sure its not the challengeTxt - if it not return it
                            keyHandles.append(base64UrlEncoded)
                    except:
                        pass
        return keyHandles

    @staticmethod
    def find_app_id(inputString):
        try:
            searchResult = re.search('"appid":"[a-z://.-_]+"', inputString).group()
            searchObject = json.loads('{' + searchResult + '}')
            return str(searchObject['appid'])
        except:
            logging.exception('Was unable to find appid value in googles SAML page')
            sys.exit(1)

    def do_login(self):
        self.session = requests.Session()
        self.session.headers['User-Agent'] = "AWS Sign-in/{} (aws-google-auth)".format(self.version)
        sess = self.get(self.login_url)

        # Collect information from the page source
        first_page = BeautifulSoup(sess.text, 'html.parser')
        gxf = first_page.find('input', {'name': 'gxf'}).get('value')
        self.cont = first_page.find('input', {'name': 'continue'}).get('value')
        page = first_page.find('input', {'name': 'Page'}).get('value')
        sign_in = first_page.find('input', {'name': 'signIn'}).get('value')
        account_login_url = first_page.find('form', {'id': 'gaia_loginform'}).get('action')

        payload = {
            'bgresponse': 'js_disabled',
            'checkConnection': '',
            'checkedDomains': 'youtube',
            'continue': self.cont,
            'Email': self.config.username,
            'gxf': gxf,
            'identifier-captcha-input': '',
            'identifiertoken': '',
            'identifiertoken_audio': '',
            'ltmpl': 'popup',
            'oauth': 1,
            'Page': page,
            'Passwd': '',
            'PersistentCookie': 'yes',
            'ProfileInformation': '',
            'pstMsg': 0,
            'sarp': 1,
            'scc': 1,
            'SessionState': '',
            'signIn': sign_in,
            '_utf8': '?',
        }

        if self.config.bg_response:
            payload['bgresponse'] = self.config.bg_response

        # GALX is sometimes not there
        try:
            galx = first_page.find('input', {'name': 'GALX'}).get('value')
            payload['GALX'] = galx
        except:
            pass

        # POST to account login info page, to collect profile and session info
        sess = self.post(account_login_url, data=payload)

        self.session.headers['Referer'] = sess.url

        # Collect ProfileInformation, SessionState, signIn, and Password Challenge URL
        challenge_page = BeautifulSoup(sess.text, 'html.parser')

        # Handle the "old-style" page
        if challenge_page.find('form', {'id': 'gaia_loginform'}):
            form = challenge_page.find('form', {'id': 'gaia_loginform'})
            passwd_challenge_url = form.get('action')
        else:
            # sometimes they serve up a different page
            logging.info("Handling new-style login page")
            form = challenge_page.find('form', {'id': 'challenge'})
            passwd_challenge_url = 'https://accounts.google.com' + form.get('action')

        for tag in form.find_all('input'):
            if tag.get('name') is None:
                continue

            payload[tag.get('name')] = tag.get('value')

        # Update the payload
        payload['Passwd'] = self.config.password

        # POST to Authenticate Password
        sess = self.post(passwd_challenge_url, data=payload)

        response_page = BeautifulSoup(sess.text, 'html.parser')
        error = response_page.find(class_='error-msg')
        cap = response_page.find('input', {'name': 'identifier-captcha-input'})

        # Were there any errors logging in? Could be invalid username or password
        # There could also sometimes be a Captcha, which means Google thinks you,
        # or someone using the same outbound IP address as you, is a bot.
        if error is not None and cap is None:
            raise ExpectedGoogleException('Invalid username or password')

        if "signin/rejected" in sess.url:
            raise ExpectedGoogleException(u'''Default value of parameter `bgresponse` has not accepted.
                Please visit login URL {}, open the web inspector and execute document.bg.invoke() in the console.
                Then, set --bg-response to the function output.'''.format(self.login_url))

        self.check_extra_step(response_page)

        # Process Google CAPTCHA verification request if present
        if cap is not None:
            self.session.headers['Referer'] = sess.url

            sess = self.handle_captcha(sess, payload)

            response_page = BeautifulSoup(sess.text, 'html.parser')
            error = response_page.find(class_='error-msg')
            cap = response_page.find('input', {'name': 'logincaptcha'})

            # Were there any errors logging in? Could be invalid username or password
            # There could also sometimes be a Captcha, which means Google thinks you,
            # or someone using the same outbound IP address as you, is a bot.
            if error is not None:
                raise ExpectedGoogleException('Invalid username or password')

            self.check_extra_step(response_page)

            if cap is not None:
                raise ExpectedGoogleException(
                    'Invalid captcha')

        self.session.headers['Referer'] = sess.url

        if "selectchallenge/" in sess.url:
            sess = self.handle_selectchallenge(sess)

        # Was there an MFA challenge?
        if "challenge/totp/" in sess.url:
            error_msg = ""
            while error_msg is not None:
                sess = self.handle_totp(sess)
                error_msg = self.parse_error_message(sess)
                if error_msg is not None:
                    logging.error(error_msg)
        elif "challenge/ipp/" in sess.url:
            sess = self.handle_sms(sess)
        elif "challenge/az/" in sess.url:
            sess = self.handle_prompt(sess)
        elif "challenge/sk/" in sess.url:
            sess = self.handle_sk(sess)
        elif "challenge/iap/" in sess.url:
            sess = self.handle_iap(sess)
        elif "challenge/ootp/5" in sess.url:
            raise NotImplementedError(
                'Offline Google App OOTP not implemented')

        # ... there are different URLs for backup codes (printed)
        # and security keys (eg yubikey) as well
        # save for later
        self.session_state = sess

    @staticmethod
    def check_extra_step(response):
        extra_step = response.find(text='This extra step shows that it’s really you trying to sign in')
        if extra_step:
            if response.find(id='contactAdminMessage'):
                raise ValueError(response.find(id='contactAdminMessage').text)

    def parse_saml(self):
        if self.session_state is None:
            # raise RuntimeError('You must use do_login() before calling parse_saml()')
            self.do_login()

        parsed = BeautifulSoup(self.session_state.text, 'html.parser')
        try:
            saml_element = parsed.find('input', {'name': 'SAMLResponse'}).get('value')
        except:

            if self.save_failure:
                logging.error("SAML lookup failed, storing failure page to "
                              "'saml.html' to assist with debugging.")
                with open("saml.html", 'w') as out:
                    out.write(str(self.session_state.text.encode('utf-8')))

            raise ExpectedGoogleException(
                'Something went wrong - Could not find SAML response, check your credentials or use --save-failure-html to debug.')

        return base64.b64decode(saml_element)

    def handle_captcha(self, sess, payload):
        response_page = BeautifulSoup(sess.text, 'html.parser')

        # Collect ProfileInformation, SessionState, signIn, and Password Challenge URL
        profile_information = response_page.find('input', {
            'name': 'ProfileInformation'
        }).get('value')
        session_state = response_page.find('input', {
            'name': 'SessionState'
        }).get('value')
        sign_in = response_page.find('input', {'name': 'signIn'}).get('value')
        passwd_challenge_url = response_page.find('form', {
            'id': 'gaia_loginform'
        }).get('action')

        # Update the payload
        payload['SessionState'] = session_state
        payload['ProfileInformation'] = profile_information
        payload['signIn'] = sign_in
        payload['Passwd'] = self.config.password

        # Get all captcha challenge tokens and urls
        captcha_container = response_page.find('div', {'id': 'identifier-captcha'})
        captcha_logintoken = captcha_container.find('input', {'id': 'identifier-token'}).get('value')
        captcha_img = captcha_container.find('div', {'class': 'captcha-img'})
        captcha_url = "https://accounts.google.com" + captcha_img.find('img').get('src')
        captcha_logintoken_audio = ''

        open_image = True

        # Check if there is a display utility installed as Image.open(f).show() do not raise any exception if not
        # if neither xv or display are available just display the URL for the user to visit.
        if os.name == 'posix' and sys.platform != 'darwin':
            if find_executable('xv') is None and find_executable('display') is None:
                open_image = False

        print("Please visit the following URL to view your CAPTCHA: {}".format(captcha_url))
        if open_image:
            try:
                with requests.get(captcha_url) as url:
                    subprocess.call(['open', url.content])

            except Exception:
                pass

        try:
            captcha_input = raw_input("Captcha (case insensitive): ") or None
        except NameError:
            captcha_input = input("Captcha (case insensitive): ") or None

        # Update the payload
        payload['identifier-captcha-input'] = captcha_input
        payload['identifiertoken'] = captcha_logintoken
        payload['identifiertoken_audio'] = captcha_logintoken_audio
        payload['checkedDomains'] = 'youtube'
        payload['checkConnection'] = 'youtube:574:1'
        payload['Email'] = self.config.username

        response = self.post(passwd_challenge_url, data=payload)

        newPayload = {}

        auth_response_page = BeautifulSoup(response.text, 'html.parser')
        print(f"\n\n\n{auth_response_page}\n\n\n\n")
        challengeId = auth_response_page.find('input', {
            'name': 'challengeId'
        }).get('value')
        challengeType = auth_response_page.find('input', {
            'name': 'challengeType'
        }).get('value')
        tl = auth_response_page.find('input', {
            'name': 'TL'
        }).get('value')
        gxf = auth_response_page.find('input', {
            'name': 'gxf'
        }).get('value')

        newPayload['challengeId'] = challengeId
        newPayload['challengeType'] = challengeType
        newPayload['TL'] = tl
        newPayload['gxf'] = gxf
        newPayload['TrustDevice'] = "on"
        newPayload['continue'] = payload['continue']
        newPayload['ltmpl'] = payload['ltmpl']
        newPayload['scc'] = payload['scc']
        newPayload['sarp'] = payload['sarp']
        newPayload['checkedDomains'] = payload['checkedDomains']
        newPayload['checkConnection'] = payload['checkConnection']
        newPayload['pstMsg'] = payload['pstMsg']
        newPayload['oauth'] = payload['oauth']
        newPayload['Email'] = payload['Email']
        newPayload['Passwd'] = payload['Passwd']

        return self.post(response.url, data=newPayload)

    def handle_sk(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        challenge_url = sess.url.split("?")[0]
        challenges_txt = response_page.find('input', {
            'name': "id-challenge"
        }).get('value')

        facet_url = urllib_parse.urlparse(challenge_url)
        facet = facet_url.scheme + "://" + facet_url.netloc

        keyHandleJSField = response_page.find('div', {'jsname': 'C0oDBd'}).get('data-challenge-ui')
        startJSONPosition = keyHandleJSField.find('{')
        endJSONPosition = keyHandleJSField.rfind('}')
        keyHandleJsonPayload = json.loads(keyHandleJSField[startJSONPosition:endJSONPosition + 1])

        keyHandles = self.find_key_handles(keyHandleJsonPayload,
                                           base64.urlsafe_b64encode(base64.b64decode(challenges_txt)))
        appId = self.find_app_id(str(keyHandleJsonPayload))

        # txt sent for signing needs to be base64 url encode
        # we also have to remove any base64 padding because including including it will prevent google accepting the auth response
        challenges_txt_encode_pad_removed = base64.urlsafe_b64encode(base64.b64decode(challenges_txt)).strip(
            '='.encode())

        u2f_challenges = [{'version': 'U2F_V2', 'challenge': challenges_txt_encode_pad_removed.decode(), 'appId': appId,
                           'keyHandle': keyHandle.decode()} for keyHandle in keyHandles]

        # Prompt the user up to attempts_remaining times to insert their U2F device.
        attempts_remaining = 5
        auth_response = None
        while True:
            try:
                auth_response_dict = u2f_auth(u2f_challenges, facet)
                auth_response = json.dumps(auth_response_dict)
                break
            except RuntimeWarning:
                logging.error("No U2F device found. %d attempts remaining",
                              attempts_remaining)
                if attempts_remaining <= 0:
                    break
                else:
                    input(
                        "Insert your U2F device and press enter to try again..."
                    )
                    attempts_remaining -= 1

        # If we exceed the number of attempts, raise an error and let the program exit.
        if auth_response is None:
            raise ExpectedGoogleException(
                "No U2F device found. Please check your setup.")

        payload = {
            'challengeId':
                response_page.find('input', {
                    'name': 'challengeId'
                }).get('value'),
            'challengeType':
                response_page.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue': response_page.find('input', {
                'name': 'continue'
            }).get('value'),
            'scc':
                response_page.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                response_page.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                response_page.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'pstMsg': '1',
            'TL':
                response_page.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                response_page.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'id-challenge':
                challenges_txt,
            'id-assertion':
                auth_response,
            'TrustDevice':
                'on',
        }
        return self.post(challenge_url, data=payload)

    def handle_sms(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        challenge_url = sess.url.split("?")[0]

        sms_token = input("Enter SMS token: G-") or None

        payload = {
            'challengeId':
                response_page.find('input', {
                    'name': 'challengeId'
                }).get('value'),
            'challengeType':
                response_page.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue':
                response_page.find('input', {
                    'name': 'continue'
                }).get('value'),
            'scc':
                response_page.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                response_page.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                response_page.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'pstMsg':
                response_page.find('input', {
                    'name': 'pstMsg'
                }).get('value'),
            'TL':
                response_page.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                response_page.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'Pin':
                sms_token,
            'TrustDevice':
                'on',
        }

        # Submit IPP (SMS code)
        return self.post(challenge_url, data=payload)

    def handle_prompt(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        challenge_url = sess.url.split("?")[0]

        data_key = response_page.find('div', {
            'data-api-key': True
        }).get('data-api-key')
        data_tx_id = response_page.find('div', {
            'data-tx-id': True
        }).get('data-tx-id')

        # Need to post this to the verification/pause endpoint
        await_url = "https://content.googleapis.com/cryptauth/v1/authzen/awaittx?alt=json&key={}".format(
            data_key)
        await_body = {'txId': data_tx_id}

        self.check_prompt_code(response_page)

        print("Open the Google App, and tap 'Yes' on the prompt to sign in ...")

        self.session.headers['Referer'] = sess.url

        retry = True
        response = None
        while retry:
            try:
                response = self.post(await_url, json=await_body)
                retry = False
            except requests.exceptions.HTTPError as ex:

                if not ex.response.status_code == 500:
                    raise ex

        parsed_response = json.loads(response.text)

        payload = {
            'challengeId':
                response_page.find('input', {
                    'name': 'challengeId'
                }).get('value'),
            'challengeType':
                response_page.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue':
                response_page.find('input', {
                    'name': 'continue'
                }).get('value'),
            'scc':
                response_page.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                response_page.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                response_page.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'checkConnection':
                'youtube:1295:1',
            'pstMsg':
                response_page.find('input', {
                    'name': 'pstMsg'
                }).get('value'),
            'TL':
                response_page.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                response_page.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'token':
                parsed_response['txToken'],
            'action':
                response_page.find('input', {
                    'name': 'action'
                }).get('value'),
            'TrustDevice':
                'on',
        }

        return self.post(challenge_url, data=payload)

    @staticmethod
    def check_prompt_code(response):
        """
        Sometimes there is an additional numerical code on the response page that needs to be selected
        on the prompt from a list of multiple choice. Print it if it's there.
        """
        num_code = response.find("div", {"jsname": "EKvSSd"})
        if num_code:
            print("numerical code for prompt: {}".format(num_code.string))

    def handle_totp(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        tl = response_page.find('input', {'name': 'TL'}).get('value')
        gxf = response_page.find('input', {'name': 'gxf'}).get('value')
        challenge_url = sess.url.split("?")[0]
        challenge_id = challenge_url.split("totp/")[1]

        mfa_token = input("MFA token: ") or None

        if not mfa_token:
            raise ValueError(
                "MFA token required for {} but none supplied.".format(
                    self.config.username))

        payload = {
            'challengeId': challenge_id,
            'challengeType': 6,
            'continue': self.cont,
            'scc': 1,
            'sarp': 1,
            'checkedDomains': 'youtube',
            'pstMsg': 0,
            'TL': tl,
            'gxf': gxf,
            'Pin': mfa_token,
            'TrustDevice': 'on',
        }

        # Submit TOTP
        return self.post(challenge_url, data=payload)

    def handle_iap(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        challenge_url = sess.url.split("?")[0]
        phone_number = input('Enter your phone number:') or None

        while True:
            try:
                choice = int(
                    input(
                        'Type 1 to receive a code by SMS or 2 for a voice call:'
                    ))
                if choice not in [1, 2]:
                    raise ValueError
            except ValueError:
                logging.error("Not a valid (integer) option, try again")
                continue
            else:
                if choice == 1:
                    send_method = 'SMS'
                elif choice == 2:
                    send_method = 'VOICE'
                else:
                    continue
                break

        payload = {
            'challengeId':
                response_page.find('input', {
                    'name': 'challengeId'
                }).get('value'),
            'challengeType':
                response_page.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue':
                self.cont,
            'scc':
                response_page.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                response_page.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                response_page.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'pstMsg':
                response_page.find('input', {
                    'name': 'pstMsg'
                }).get('value'),
            'TL':
                response_page.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                response_page.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'phoneNumber':
                phone_number,
            'sendMethod':
                send_method,
        }

        # Submit phone number and desired method (SMS or voice call)
        sess = self.post(challenge_url, data=payload)

        response_page = BeautifulSoup(sess.text, 'html.parser')
        challenge_url = sess.url.split("?")[0]

        token = input("Enter " + send_method + " token: G-") or None

        payload = {
            'challengeId':
                response_page.find('input', {
                    'name': 'challengeId'
                }).get('value'),
            'challengeType':
                response_page.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue':
                response_page.find('input', {
                    'name': 'continue'
                }).get('value'),
            'scc':
                response_page.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                response_page.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                response_page.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'pstMsg':
                response_page.find('input', {
                    'name': 'pstMsg'
                }).get('value'),
            'TL':
                response_page.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                response_page.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'pin':
                token,
        }

        # Submit SMS/VOICE token
        return self.post(challenge_url, data=payload)

    def handle_selectchallenge(self, sess):
        response_page = BeautifulSoup(sess.text, 'html.parser')
        # Known mfa methods, 5 is disabled till its implemented
        auth_methods = {
            2: 'TOTP (Google Authenticator)',
            3: 'SMS',
            4: 'OOTP (Google Prompt)'
            # 5: 'OOTP (Google App Offline Security Code)'
        }

        unavailable_challenge_ids = [
            int(i.attrs.get('data-unavailable'))
            for i in response_page.find_all(
                lambda tag: tag.name == 'form' and 'data-unavailable' in tag.attrs
            )
        ]

        # ootp via google app offline code isn't implemented. make sure its not valid.
        unavailable_challenge_ids.append(5)

        challenge_ids = [
            int(i.get('value'))
            for i in response_page.find_all('input', {'name': 'challengeId'})
            if int(i.get('value')) not in unavailable_challenge_ids
        ]

        challenge_ids.sort()

        auth_methods = {
            k: auth_methods[k]
            for k in challenge_ids
            if k in auth_methods and k not in unavailable_challenge_ids
        }

        print('Choose MFA method from available:')
        print('\n'.join(
            '{}: {}'.format(*i) for i in list(auth_methods.items())))

        selected_challenge = input("Enter MFA choice number ({}): ".format(
            challenge_ids[-1:][0])) or None

        if selected_challenge is not None and int(selected_challenge) in challenge_ids:
            challenge_id = int(selected_challenge)
        else:
            # use the highest index as that will default to prompt, then sms, then totp, etc.
            challenge_id = challenge_ids[-1:][0]

        print("MFA Type Chosen: {}".format(auth_methods[challenge_id]))

        # We need the specific form of the challenge chosen
        challenge_form = response_page.find(
            'form', {'data-challengeentry': challenge_id})

        payload = {
            'challengeId':
                challenge_id,
            'challengeType':
                challenge_form.find('input', {
                    'name': 'challengeType'
                }).get('value'),
            'continue':
                challenge_form.find('input', {
                    'name': 'continue'
                }).get('value'),
            'scc':
                challenge_form.find('input', {
                    'name': 'scc'
                }).get('value'),
            'sarp':
                challenge_form.find('input', {
                    'name': 'sarp'
                }).get('value'),
            'checkedDomains':
                challenge_form.find('input', {
                    'name': 'checkedDomains'
                }).get('value'),
            'pstMsg':
                challenge_form.find('input', {
                    'name': 'pstMsg'
                }).get('value'),
            'TL':
                challenge_form.find('input', {
                    'name': 'TL'
                }).get('value'),
            'gxf':
                challenge_form.find('input', {
                    'name': 'gxf'
                }).get('value'),
            'subAction':
                challenge_form.find('input', {
                    'name': 'subAction'
                }).get('value'),
        }
        if challenge_form.find('input', {'name': 'SendMethod'}) is not None:
            payload['SendMethod'] = challenge_form.find(
                'input', {
                    'name': 'SendMethod'
                }).get('value')

        # POST to google with the chosen challenge
        return self.post(
            self.base_url + challenge_form.get('action'), data=payload)



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DevExport(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(pexpect.spawn(f'{CLI_NAME} {Utils.get_first(iam)} {Utils.get_first(export)} '
                                    f'--env {DEFAULT_ENV} --skip-upgrade {extra_args}', timeout=5, encoding='utf-8'),
                                    extra_args=extra_args)
        self._child.delayafterread = .01
        self._child.delaybeforesend = .5

    def run(self):
        self.step("Testing IAM export.")
        self.export()

    def export(self):
        print(f"Testing IAM export.")
        self.expect('.*Successfully updated.*')
        print("`figgy iam export` passed.")



from sty import fg

from figcli.config.style.color import Color
from figcli.config.style.palette import Palette
from figcli.config.style.terms.term import Term


class AppleTerminal(Term):
    TERM_PROGRAM = 'Apple_Terminal'
    BLUE = (33, 46, 222)
    GREEN = (40, 138, 74)
    RED = (191, 29, 42)
    YELLOW = (199, 142, 10)

    def get_colors(self) -> Color:
        return Color(
            colors_enabled=self.colors_enabled,
            palette=Palette(
                blue=AppleTerminal.BLUE,
                green=AppleTerminal.GREEN,
                red=AppleTerminal.RED,
                yellow=AppleTerminal.YELLOW
            )
        )



import re
import xml.etree.ElementTree as ET

from figcli.config.sso import *
from figcli.config.constants import GOOGLE_SESSION_CACHE_PATH
from dataclasses import dataclass
from typing import Optional, Any, List

from figcli.models.assumable_role import AssumableRole
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.auth.google.google import Google, ExpectedGoogleException

from figcli.models.defaults.defaults import CLIDefaults
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.auth.provider.sso_session_provider import SSOSessionProvider
from figcli.svcs.vault import FiggyVault
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils
from figcli.config.constants import ERROR_LOG_DIR, DISABLE_KEYRING

@dataclass
class GoogleConfig:
    username: str
    password: str
    idp_id: str
    sp_id: str
    bg_response: Optional[str]


class GoogleSessionProvider(SSOSessionProvider):
    _SESSION_CACHE_KEY = 'session'
    _SAML_CACHE_KEY = 'saml_assertion'
    _SAML_MAX_AGE = 60 * 60 * 8 * 1000  # 12 hours

    def __init__(self, defaults: CLIDefaults):
        super().__init__(defaults)
        keychain_enabled = defaults.extras.get(DISABLE_KEYRING) is not True
        vault = FiggyVault(keychain_enabled=keychain_enabled)
        self._cache_manager: CacheManager = CacheManager(file_override=GOOGLE_SESSION_CACHE_PATH, vault=vault)
        config = GoogleConfig(
            username=defaults.user,
            password=SecretsManager.get_password(defaults.user),
            idp_id=defaults.provider_config.idp_id,
            sp_id=defaults.provider_config.sp_id,
            bg_response=None)

        self._google = Google(config=config, save_failure=False)
        self._write_google_session_to_cache(self._google.session_state)

    def _write_google_session_to_cache(self, session: Any):
        self._cache_manager.write(self._SESSION_CACHE_KEY, session)

    def get_assumable_roles(self) -> List[AssumableRole]:
        return self._cache_manager.get_val_or_refresh('assumable_roles', refresher=self.__lookup_roles)

    def __lookup_roles(self) -> List[AssumableRole]:
        try:
            saml = self._google.parse_saml()
        except ExpectedGoogleException as e:
            self._utils.error_exit(e)
        else:
            decoded_assertion = saml.decode('utf-8')
            root = ET.fromstring(decoded_assertion)

            prefix_map = {"saml2": "urn:oasis:names:tc:SAML:2.0:assertion"}
            role_attribute = root.find(".//saml2:Attribute[@Name='https://aws.amazon.com/SAML/Attributes/Role']",
                                       prefix_map)

            # SAML arns should look something like this:
            # arn:aws:iam::9999999999:role/figgy-dev-devops,arn:aws:iam::9999999999:saml-provider/google
            pattern = r'^arn:aws:iam::([0-9]+):role/(\w+-(\w+)-(\w+)),.*saml-provider/(\w+)'
            assumable_roles: List[AssumableRole] = []
            for value in role_attribute.findall('.//saml2:AttributeValue', prefix_map):
                result = re.search(pattern, value.text)
                unparsable_msg = f'{value.text} is of an invalid pattern, it must match: {pattern} for figgy to ' \
                                 f'dynamically map account_id -> run_env -> role for Google users. Are you sure that' \
                                 f'you mapped your SAML Attributes correctly in Google Admin Console? The decoded SAML ' \
                                 f'assertion that caused the error has been saved here: {ERROR_LOG_DIR}/saml-error.xml'
                if not result:
                    Utils.write_error('saml-error.xml', unparsable_msg)
                    Utils.stc_error_exit(unparsable_msg)

                result.groups()
                account_id, role_name, run_env, role, provider_name = result.groups()

                if not account_id or not run_env or not role_name or not role:
                    Utils.write_error('saml-error.xml', decoded_assertion)
                    Utils.stc_error_exit(unparsable_msg)
                else:
                    assumable_roles.append(AssumableRole(account_id=account_id,
                                                         role=Role(role, full_name=role_name),
                                                         run_env=RunEnv(run_env),
                                                         provider_name=provider_name,
                                                         profile=None))

            return assumable_roles

    def _get_decoded_saml(self) -> str:
        self._google.do_login()
        return self._google.parse_saml().decode('utf-8')

    def get_saml_assertion(self, prompt: bool = False) -> str:
        return self._cache_manager.get_val_or_refresh(SAML_ASSERTION_CACHE_KEY, self._get_decoded_saml,
                                                      max_age=SAML_ASSERTION_CACHE_KEY)

    def cleanup_session_cache(self):
        self._write_google_session_to_cache(None)



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.dev.get import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DataDelete(FiggyTest):
    def __init__(self, extra_args=""):
        print(f"Testing `{CLI_NAME} config {Utils.get_first(delete)} --env {DEFAULT_ENV}`")
        super().__init__(pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(delete)} --env {DEFAULT_ENV}'
                                       f' --skip-upgrade {extra_args}',
                                    timeout=5, encoding='utf-8'), extra_args=extra_args)

    def run(self):
        self.step(f"Testing delete for param: {param_1}")
        self.delete(param_1)

    def delete(self, name, delete_another=False, repl_source_delete=False, repl_dest_delete=False):
        self.expect('.*PS Name.*')
        self.sendline(name)
        print(f"Delete sent for {name}")
        if repl_source_delete:
            self.expect(".*You may NOT delete sources that are actively replicating.*")
        elif repl_dest_delete:
            match = self.expect_multiple([f".*active replication destination.*{name}.*", ".*deleted successfully.*"])
            print(f"Match: {match}")
            if match == 0:
                print("Matched replication destination message")
                self.sendline('y')
                self.expect(f".*{name}.*deleted successfully.*Delete another.*")
            else:
                print("Matched delete successful message")
        else:
            self.expect(f'.*deleted successfully.*Delete another.*')
            print("Validating delete success.")
            get = DevGet(extra_args=self.extra_args)
            get.get(param_1, param_1_val, expect_missing=True)

        if delete_another:
            self.sendline('y')
        else:
            self.sendline('n')

        print("Successful delete validated.")




DEFAULT_SESSION_DURATION = 43200  # 12 hours
SANDBOX_SESSION_DURATION = 60 * 60 # 1 hour

BASTION_PROFILE_ENV_NAME = 'FIGGY_AWS_PROFILE'

AWS_REGIONS = ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2', 'af-south-1', 'ap-east-1', 'ap-east-2',
               'ap-northeast-3', 'ap-northeast-2', 'ap-northeast-1', 'ap-southeast-1', 'ap-southeast-2',
               'ca-central-1', 'cn-north-1', 'cn-northwest-1', 'eu-central-1', 'eu-west-1', 'eu-west-2',
               'eu-west-3', 'eu-north-1', 'me-south-1', 'sa-east-1', 'us-gov-east-1', 'us-gov-west-1']
# Colors:
green = 'green'
blue = 'blue'
red = 'red'



from figcli.commands.factory import Factory
from figcli.commands.iam.export import Export
from figcli.commands.iam_context import IAMContext
from figcli.utils.utils import *
from figcli.config import *
from boto3.session import Session
from typing import Optional


class IAMFactory(Factory):
    def __init__(self, command: frozenset, context: IAMContext, env_session: Session,
                 all_sessions: Optional[Dict[str, Session]] = None):

        self._all_sessions: Optional[Dict[str, Session]] = all_sessions
        self._command = command
        self._utils = Utils(context.colors_enabled)
        self._iam_context: IAMContext = context
        self._env_session: Session = env_session

    def instance(self):
        return self.get(self._command)

    def get(self, command: frozenset):
        if command == export:
            return Export(self._iam_context, self._env_session, self._all_sessions)
        else:
            self._utils.error_exit(f"{command} is not a valid IAM command. You must select from: "
                                   f"[{CollectionUtils.printable_set(iam_commands)}]. Try using --help for more info.")



import logging
import requests
import os
import json
import traceback
import datetime
import re
from figcli.config import *
from figcli.config.style.terminal_factory import TerminalFactory
from figcli.input import Input
from figcli.utils.utils import Utils

from figcli.models.defaults.defaults import CLIDefaults

log = logging.getLogger(__name__)


def obey_reporting_preference(function):
    """
    Do not execute this method if report_errors is disabled, even if someone tells you to!
    """

    def inner(self, *args, **kwargs):
        if not self.reporting_enabled:
            return None
        else:
            return function(self, *args, **kwargs)

    return inner


class FiggyErrorReporter:
    _SECRET_MATCHERS = [
        r'.*AKIA.*',  # Match for AWS access keys
        r'.*[pP][aA][sS][sS].*'  # Anything with pass(word), skip it.
    ]

    def __init__(self, cli_defaults: CLIDefaults):
        self._cli_defaults = cli_defaults
        self.c = TerminalFactory(self._cli_defaults.colors_enabled).instance().get_colors()
        self.reporting_enabled = cli_defaults.report_errors

    @staticmethod
    def sanitize(e: Exception):
        """
        FYI: There's no reason why there _should_ ever be a secret in a stack trace, but.. if.. somehow.. there is..
             then we want to abandon ship!

        Look for potential stack trace and wipe this baby clean if anything matches.
        :param e: Exception to sanitize and return a printable stack trace of
        :return: printable representation of the provided exception
        """
        printable_exception = ''.join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__))

        for matcher in FiggyErrorReporter._SECRET_MATCHERS:
            if re.match(matcher, printable_exception, re.MULTILINE):
                return f"Stacktrace wiped because matcher: {matcher} matched it. "

        return printable_exception

    @obey_reporting_preference
    def report_error(self, command: str, e: Exception) -> None:
        """
        If the user chooses to report this exception, ship it off over to the Figgy API to let us know what went wrong!
        :param command: user's command input
        :param e: exception that was thrown
        """

        printable_exception = self.sanitize(e)
        os = Utils.get_os()

        payload = {
            'command': command,
            'os': os,
            'stacktrace': printable_exception
        }

        result = requests.post(FIGGY_ERROR_REPORTING_URL, json=payload)

        Utils.stc_validate(result.status_code == 200, "Unable to report this error to Figgy. Please consider "
                                                      f"opening a ticket on the figgy github repo: {FIGGY_GITHUB}")

        print(f"We are so sorry you experienced this error! This error has been anonymously reported to the Figgy "
              f"development team. \n\nIf you don't want to be prompted to report errors, you can disable the error "
              f"reporting by running `{CLI_NAME} --configure`.")
        print(f"\n\n{self.c.fg_bl}--------------------------------------------------------{self.c.rs}\n\n")

    def log_error(self, command: str, e: Exception):
        os.makedirs(ERROR_LOG_DIR, exist_ok=True)

        # Do not sanitize here since this is only written to the user's local log folder.
        printable_exception = ''.join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__))
        log_file_name = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S.log")

        with open(f'{ERROR_LOG_DIR}/{log_file_name}', "w+") as log:
            log.write(f'Command: {command}\n\n\n{printable_exception}')

        print(f"\n\n{self.c.fg_bl}---------------------ERROR ENCOUNTERED --------------------{self.c.rs}\n\n")
        print(f"{self.c.fg_rd}Figgy experienced the following irrecoverable error. "
              f"Please consider reporting this error.{self.c.rs}")
        print(printable_exception)

        print(f"\n\n{self.c.fg_bl}-----------------------------------------------------------{self.c.rs}\n\n")

        if self.reporting_enabled:
            ship_it = Input.y_n_input("Would you like to report this to Figgy's developer(s)?", default_yes=True)

            if ship_it:
                self.report_error(command, e)
            else:
                print(f"Error was not reported. Please consider reporting this error on our Github: "
                      f"{FIGGY_GITHUB}. Error details have been logged to this file: "
                      f"{self.c.fg_bl}{log_file_name}{self.c.rs}. Farewell! \n")



from typing import Union

from figcli.commands.factory import Factory
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.defaults.provider import Provider
from figcli.svcs.auth.google.google_session_provider import GoogleSessionProvider
from figcli.svcs.auth.bastion.bastion_session_provider import BastionSessionProvider
from figcli.svcs.auth.okta.okta_session_provider import OktaSessionProvider
from figcli.svcs.auth.profile.profile_session_provider import ProfileSessionProvider
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.svcs.auth.provider.sso_session_provider import SSOSessionProvider


class SessionProviderFactory(Factory):

    def __init__(self, defaults: CLIDefaults):
        self._defaults = defaults
        self.__bastion_session_provider = None
        self.__okta_session_provider = None
        self.__google_session_provider = None
        self.__profile_provider = None

    def instance(self) -> Union[SSOSessionProvider, SessionProvider]:
        if self._defaults.provider is Provider.OKTA:
            if not self.__okta_session_provider:
                self.__okta_session_provider = OktaSessionProvider(self._defaults)

            return self.__okta_session_provider
        elif self._defaults.provider is Provider.AWS_BASTION:
            if not self.__bastion_session_provider:
                self.__bastion_session_provider = BastionSessionProvider(self._defaults)

            return self.__bastion_session_provider
        elif self._defaults.provider is Provider.GOOGLE:
            if not self.__google_session_provider:
                self.__google_session_provider = GoogleSessionProvider(self._defaults)

            return self.__google_session_provider
        elif self._defaults.provider is Provider.PROFILE:
            if not self.__profile_provider:
                self.__profile_provider = ProfileSessionProvider(self._defaults)

            return self.__profile_provider
        else:
            raise NotImplementedError(f"Provider: {self._defaults.provider} is not currently supported.")



from enum import Enum

"""
The below classes represent a figgy config file found under ~/.figgy/config that provides defaults and override values
so users don't have to go through the full configuration workflow when running `figgy --configure`
"""


class ConfigSection:
    NAME = None


class Config:
    class Section:
        class Okta(ConfigSection, Enum):
            NAME = 'OKTA'
            APP_LINK = 'app_link'
            FACTOR_TYPE = 'factor_type'

        class Google(ConfigSection, Enum):
            NAME = 'GOOGLE'
            IDP_ID = 'identity_provider_id'
            SP_ID = 'service_provider_id'

        class Bastion(ConfigSection, Enum):
            NAME = 'BASTION'
            PROFILE = 'profile'

        class Figgy(ConfigSection, Enum):
            NAME = 'FIGGY'
            MFA_ENABLED = 'mfa_enabled'
            AUTO_MFA = 'auto_mfa'
            COLORS_ENABLED = 'colors_enabled'
            REPORT_ERRORS = 'report_errors'
            AWS_REGION = 'aws_region'
            USAGE_TRACKING = 'anonymous_metrics_enabled'


FAKE_GOOGLE_IDP_ID = 'N0tre4le3'
FAKE_GOOGLE_SP_ID = '12345678010'
FAKE_OKTA_APP_LINK = 'https://your-domain.okta.com/home/amazon_aws/AsdF1afak123145faf1/123'

EMPTY_CONFIG = f"""
[FIGGY]
mfa_enabled = false
auto_mfa = false
colors_enabled = true
report_errors = true
anonymous_metrics_enabled = true
aws_region = us-east-1

[OKTA]
app_link = {FAKE_OKTA_APP_LINK}
factor_type = GOOGLE

[GOOGLE]
identity_provider_id = {FAKE_GOOGLE_IDP_ID}
service_provider_id = {FAKE_GOOGLE_SP_ID}

[BASTION]
profile = bastion
"""



from sty import fg

from figcli.config.style.palette import Palette


class Color:
    def __init__(self, colors_enabled: bool, palette=Palette()):
        self.bl, self.gr, self.rd, self.fg_bl, self.fg_gr, self.fg_rd, self.rs = '', '', '', '', '', '', ''
        self.yl, self.fg_yl, self.bl_val, self.gr_val, self.yl_val, self.rd_val = '', '', '', '', '', ''
        if colors_enabled:
            self.bl = 'blue'
            self.rd = 'red'
            self.yl = 'yellow'
            self.gr = 'green'

            self.bl_val = palette.bl
            self.gr_val = palette.gr
            self.rd_val = palette.rd
            self.yl_val = palette.yl

            self.fg_bl = palette.fg_bl
            self.fg_gr = palette.fg_gr
            self.fg_rd = palette.fg_rd
            self.fg_yl = palette.fg_yl
            self.rs = fg.rs



import time


class AuditLog:
    """
    Represents a singular audit log retrieved from our config-auditor table.
    """
    def __init__(self, parameter_name: str, time: int, action: str, user: str):
        self.parameter_name = parameter_name
        self.time = time
        self.action = action
        self.user = user

    def __str__(self):
        return f"Parameter: {self.parameter_name}\n" \
               f"Time: {time.ctime(int(self.time / 1000))}\n" \
               f"User: {self.user}\n" \
               f"Action: {self.action}\n"
